/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import java.util.*;

/**
 * Model interface
 * 
 * @author Szymon Książek
 *
 */
public interface ARDModel 
{
	/**
	 * Returns list of all attributes
	 * 
	 * @return		list of all attributes 
	 */
	public List<Attribute> getAttributesList();
	
	
	/**
	 * Sets list of attributes
	 * 
	 * @param attributesList	list of new attributes
	 */
	public void setAttributesList(List<Attribute> attributesList);
	
	
	/**
	 * Returns list of all properties
	 * 
	 * @return		list of all properties
	 */
	public List<Property> getPropertiesList();
	
	
	/**
	 * Sets list of properties
	 * 
	 * @param propertiesList	list of new properties
	 */
	public void setPropertiesList(List<Property> propertiesList);
	
	
	/**
	 * Returns list of all transforms
	 * 
	 * @return		list of all transforms
	 */
	public List<Transform> getTransformsList();
	
	
	/**
	 * Sets list of transforms
	 * 
	 * @param transformsList	list of new transforms 
	 */
	public void setTransformsList(List<Transform> transformsList);
	
	
	/**
	 * Returns list of all dependencies
	 * 
	 * @return		list of all dependencies
	 */
	public List<Dependency> getDependenciesList();
	
	
	/**
	 * Sets list of dependencies
	 * 
	 * @param dependenciesList	list of new dependencies
	 */
	public void setDependenciesList(List<Dependency> dependenciesList);
	
	
	/**
	 * Returns list of properties which are independent for given property
	 * 
	 * @param dependentProperty		
	 * 
	 * @return		list of independent properties
	 */
	public List<Property> getIndependentProperties(Property dependentProperty);
	
	
	/**
	 * Returns list of properties which are dependent for given property
	 * 
	 * @param independentProperty
	 * 
	 * @return		list of dependent properties
	 */
	public List<Property> getDependentProperties(Property independentProperty);
	
	
	/**
	 * Returns list of sources for given property
	 * 
	 * @param destinyProperty
	 * 
	 * @return		list of sources properties
	 */
	public List<Property> getTransformSources(Property destinyProperty);
	
	
	/**
	 * Returns list of destinies for given property
	 * 
	 * @param sourceProperty
	 * 
	 * @return		list of destiny properties
	 */
	public List<Property> getTransformDestinies(Property sourceProperty);
	
	
	/**
	 * Adds attribute witch given name to model
	 * 
	 * @param attributeName	attribute name
	 * 
	 * @return				added attribute	
	 */
	public Attribute addAttribute(String attributeName);
	
	
	/**
	 * Returns an attribute witch given name
	 * 
	 * @param attributeName	attribute name
	 * 
	 * @return				attribute witch given name
	 */
	public Attribute getAttribute(String attributeName);
	
	
	/**
	 * Adds property consists of given attributes to model 
	 * 
	 * @param attributes	propety attributes 
	 * 
	 * @return				added property
	 */
	public Property addProperty(List<Attribute> attributes);
	
	
	/**
	 * Adds finalization to model
	 * 
	 * @param property			finalized property
	 * 
	 * @param attributesNames	list of attributes names of new property
	 * 
	 * @return					property created by finalization operation
	 */
	public Property addFinalization(Property property, List<String> attributesNames);
	
	
	/**
	 * Adds split to model
	 * 
	 * @param property			source of split
	 * 
	 * @param newProperties		list of properties created by split operation
	 * 
	 * @param newDependencies	list of dependencies created by split operation
	 */
	public void addSplit(Property property, List<Property> newProperties, List<Dependency> newDependencies);
	
	
	/**
	 * Changes attribute name
	 * 
	 * @param oldName	old name of attribute
	 * 
	 * @param newName	new name of	attribute
	 */
	public void changeAttributeName(String oldName, String newName);
	
	
	/**
	 * Removes attribute from model
	 * 
	 * @param removedAttribute		removed attribute
	 */
	public void removeAttribute(Attribute removedAttribute);
	
	
	/**
	 * Adds new attribute to given property
	 * 
	 * @param property			property which is given a new attribute
	 * 
	 * @param attributeName		name of new attribute
	 */
	public void addAttributeToProperty(Property property, String attributeName);
	
	
	/**
	 * Checks if dependency between given properties is allowed.
	 * 
	 * @param firstProperty		first property
	 * @param secondProperty	second property
	 * 
	 * @return		-1 if dependency between first and second property is allowed, 1 if dependency between
	 * 				second and first property is allowed, 0 if dependency between given properties is not allowed
	 */
	public int isDependencyAllowed(Property firstProperty, Property secondProperty);
}
