/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import java.util.*;

/**
 * Class that implements functionality of model part of application
 * 
 * @author Szymon Książek
 *
 */
public class ARDModelImpl implements ARDModel 
{
	/**
	 * List of attributes
	 */
	private List<Attribute> attributesList = null;
	
	/**
	 * List of properties
	 */
	private List<Property> propertiesList = null;
	
	/**
	 * List of transforms
	 */
	private List<Transform> transformsList = null;
	
	/**
	 * List of dependencies
	 */
	private List<Dependency> dependenciesList = null;
	
	
	/**
	 * Constructor 
	 */
	public ARDModelImpl()
	{
		attributesList = new ArrayList<Attribute>();
		propertiesList = new ArrayList<Property>();
		transformsList = new ArrayList<Transform>();
		dependenciesList = new ArrayList<Dependency>();
	}


	public List<Attribute> getAttributesList() 
	{
		return attributesList;
	}
	
	
	public void setAttributesList(List<Attribute> attributesList) 
	{
		this.attributesList = attributesList;
	}
	
	
	public List<Property> getPropertiesList() 
	{
		return propertiesList;
	}
	
	
	public void setPropertiesList(List<Property> propertiesList) 
	{
		this.propertiesList = propertiesList;
	}


	public List<Transform> getTransformsList() 
	{
		return transformsList;
	}


	public void setTransformsList(List<Transform> transformsList) 
	{
		this.transformsList = transformsList;
	}


	public List<Dependency> getDependenciesList() 
	{
		return dependenciesList;
	}


	public void setDependenciesList(List<Dependency> dependenciesList) 
	{
		this.dependenciesList = dependenciesList;
		
		this.clearDependencies();
	}
	
	
	public List<Property> getIndependentProperties(Property dependentProperty)
	{
		List<Property> independentProperties = new ArrayList<Property>();
		
		List<Dependency> allDependenciesList = getAllDependenciesList();
		
		for (Dependency dependency : allDependenciesList)
		{
			if (dependency.getDependent() == dependentProperty &&
					!independentProperties.contains(dependency.getIndependent()))
			{
				independentProperties.add(dependency.getIndependent());
			}
		}
		
		return independentProperties;
	}
	
	
	public List<Property> getDependentProperties(Property independentProperty)
	{
		List<Property> dependentProperties = new ArrayList<Property>();
		
		List<Dependency> allDependenciesList = getAllDependenciesList();
		
		for (Dependency dependency : allDependenciesList)
		{
			if (dependency.getIndependent() == independentProperty &&
					!dependentProperties.contains(dependency.getDependent()))
			{
				dependentProperties.add(dependency.getDependent());
			}
		}
		
		return dependentProperties;
	}
	
	
	public List<Property> getTransformSources(Property destinyProperty)
	{
		List<Property> transformSources = new ArrayList<Property>();
		
		for (Transform transform : transformsList)
		{
			if (transform.getDestiny() == destinyProperty &&
					!transformSources.contains(transform.getSource()))
			{
				transformSources.add(transform.getSource());
			}
		}
		
		return transformSources;
	}
	
	
	public List<Property> getTransformDestinies(Property sourceProperty)
	{
		List<Property> transformDestinies = new ArrayList<Property>();
		
		for (Transform transform : transformsList)
		{
			if (transform.getSource() == sourceProperty &&
					! transformDestinies.contains(transform.getDestiny()))
			{
				transformDestinies.add(transform.getDestiny());
			}
		}
		
		return transformDestinies;
	}
	
	
	public Attribute addAttribute(String attributeName)
	{
		Attribute attribute = new Attribute(attributeName.trim());
		
		attributesList.add(attribute);
		
		return attribute;
	}
	
	
	public Attribute getAttribute(String attributeName)
	{
		Attribute foundAttribute = null;
		
		for (Attribute attribute : attributesList)
		{
			if (attribute.getName().equals(attributeName.trim()))
			{
				foundAttribute = attribute;
				break;
			}
		}
		
		return foundAttribute;
	}
	
	
	public Property addProperty(List<Attribute> attributes)
	{
		Property property = new Property(attributes);
		
		propertiesList.add(property);
		
		return property;
	}
	
	
	public Property addFinalization(Property property, List<String> attributesNames)
	{
		List<Dependency> inputDependencies = getInputDependencies(property);
		List<Dependency> outputDependencies = getOutputDependencies(property);
		
		removeDownFromProperty(property, false);
		
		// create new property and add it to model
		List<Attribute> newAttributes = new ArrayList<Attribute>();
		
		Property newProperty = null;
		
		for (String newAttributeName : attributesNames)
		{
			newAttributes.add(addAttribute(newAttributeName.trim()));
		}
		
		newProperty = addProperty(newAttributes);
		
		// operations of TPH diagram
		addTransform(property, newProperty);
		
		// change in dependencies finalized property to new property
		for (Dependency inputDependency : inputDependencies)
		{
			addDependency(inputDependency.getIndependent(), newProperty);
		}
		
		for (Dependency outputDependency : outputDependencies)
		{
			addDependency(newProperty, outputDependency.getDependent());
		}
		
		// remove all not necessary dependencies
		clearDependencies();

		return newProperty;
	}
	
	
	public void addSplit(Property property, List<Property> newProperties, List<Dependency> newDependencies)
	{
		List<Dependency> inputDependencies = getInputDependencies(property);
		List<Dependency> outputDependencies = getOutputDependencies(property);
		
		removeDownFromProperty(property, false);
	
		propertiesList.addAll(newProperties);
		
		// for each one of new properties add transform on TPH diagram
		for (Property newProperty : newProperties)
		{
			addTransform(property, newProperty);
		}
		
		// add to model dependencies which were created in split dialog
		for (Dependency dependency : newDependencies)
		{
			addDependency(dependency.getIndependent(), dependency.getDependent());
		}
			
		// remove all dependencies that uses old property, because that property isn't visible on ARG graph
		dependenciesList.removeAll(inputDependencies);
		dependenciesList.removeAll(outputDependencies);
		
		// remove all not necessary dependencies
		clearDependencies();
	}
	
	
	public void changeAttributeName(String oldName, String newName)
	{
		for (Attribute attribute : attributesList)
		{
			if (attribute.getName().equals(oldName.trim()))
			{
				attribute.setName(newName.trim());
			}
		}
	}
	
	
	
	public void removeAttribute(Attribute removedAttribute)
	{
		List<Attribute> removedAttributesList = new ArrayList<Attribute>();
		List<Property> removedPropertiesList = new ArrayList<Property>();
		
		
		// pierwsza w TPH wlaściwość w której pojawił się dany atrybut
		Property firstOccuranceProperty = null;
		
		for (Transform transform : getTransformsList())
		{
			if (!transform.getSource().getAttributes().contains(removedAttribute) &&
					transform.getDestiny().getAttributes().contains(removedAttribute))
			{
				firstOccuranceProperty = transform.getDestiny();
				break;
			}
		}
		
		
		// jeśli nie znaleziono tzn. że usuwamy korzeń drzewa, szukamy korzenia
		if (firstOccuranceProperty == null)
		{
			for (Property property : getPropertiesList())
			{
				if (property.getAttributes().size() == 1 && property.getAttributes().contains(removedAttribute))
				{
					firstOccuranceProperty = property;
					break;
				}
			}
		}
		
		// w tym momencie w zmiennej firstOccuranceProperty mamy pierwszą property w której pojawił się usuwany atrybut
		removedAttributesList.add(removedAttribute);
		
		recurentRemoveAttribute(firstOccuranceProperty, removedAttributesList, removedPropertiesList);
		
		
		getAttributesList().removeAll(removedAttributesList);
		getPropertiesList().removeAll(removedPropertiesList);
		
		clearTransforms();
		clearDependencies();
	}
	
	
	public int isDependencyAllowed(Property firstProperty, Property secondProperty)
	{
		if (firstProperty == secondProperty && 
				(getTransformDestinies(firstProperty).size() > 0 || getTransformDestinies(secondProperty).size() > 0))
		{
			return 0;
		}
		
		List<Property> firstPropertyLeafs = new ArrayList<Property>();
		List<Property> secondPropertyLeafs = new ArrayList<Property>();
		
		recurentFindPropertyLeafs(firstProperty, firstPropertyLeafs);
		recurentFindPropertyLeafs(secondProperty, secondPropertyLeafs);
		
		for (Dependency dependency : getDependenciesList())
		{
			if (firstPropertyLeafs.contains(dependency.getIndependent()) &&
					secondPropertyLeafs.contains(dependency.getDependent()))
			{
				return -1;
			}
			
			if (secondPropertyLeafs.contains(dependency.getIndependent()) &&
					firstPropertyLeafs.contains(dependency.getDependent()))
			{
				return 1;
			}
		}
		
		return 0;
	}
	
	
	public void addAttributeToProperty(Property property, String attributeName)
	{
		Attribute newAttribute = this.addAttribute(attributeName);
		
		recurentAddAttributeToProperty(property, newAttribute);
	}
	
	
	private void recurentRemoveAttribute(Property actualProperty, 
			List<Attribute> removedAttributesList, 
			List<Property> removedPropertiesList)
	{
		actualProperty.getAttributes().removeAll(removedAttributesList);
		
		if (actualProperty.getAttributes().size() == 0)
		{
			Property independentProperty = null;
			
			for (Transform transform : getTransformsList())
			{
				if (transform.getDestiny() == actualProperty)
				{
					independentProperty = transform.getSource();
					break;
				}
			}
			
			if (independentProperty != null)
			{
				List<Property> removedPropertyLeafs = new ArrayList<Property>();
				
				recurentFindPropertyLeafs(actualProperty, removedPropertyLeafs);
				
				for (Dependency dependency : getDependenciesList())
				{
					if (removedPropertyLeafs.contains(dependency.getIndependent()))
					{
						dependency.setIndependent(independentProperty);
					}
					if (removedPropertyLeafs.contains(dependency.getDependent()))
					{
						
						dependency.setDependent(independentProperty);
					}
				}
			}
			
			removedPropertiesList.add(actualProperty);
			
			if (getTransformDestinies(actualProperty) != null && getTransformDestinies(actualProperty).size() == 1)
			{
				for (Property transformDestiny : getTransformDestinies(actualProperty))
				{
					removedAttributesList.addAll(transformDestiny.getAttributes());
				}
			}
		}
		
		for (Property transformDestiny : getTransformDestinies(actualProperty))
		{
			recurentRemoveAttribute(transformDestiny, removedAttributesList, removedPropertiesList);
		}
	}
	
	
	/**
	 * Creates list of properties that are leafs created from given property
	 * 
	 * @param property			property
	 * @param propertyLeafs		leafs created from given property
	 */
	private void recurentFindPropertyLeafs(Property property, List<Property> propertyLeafs)
	{
		List<Property> transformDestiniesList = getTransformDestinies(property);
		
		if (transformDestiniesList.size() == 0)
		{
			propertyLeafs.add(property);
		}
		else
		{
			for (Property destinyProperty : transformDestiniesList)
			{
				recurentFindPropertyLeafs(destinyProperty, propertyLeafs);
			}
		}
	}
	
	
	/**
	 * Returns list of dependencies which has given property as independent property. Used locally.
	 * 
	 * @param property 	independent property
	 * 
	 * @return 			dependencies which has given property as independent property
	 */
	private List<Dependency> getInputDependencies(Property property)
	{
		ArrayList<Dependency> inputDependencies = new ArrayList<Dependency>();
		
		for (Dependency dependency : getAllDependenciesList())
		{
			if (dependency.getDependent() == property)
			{
				inputDependencies.add(dependency);
			}
		}
		
		return inputDependencies;
	}
	
	
	/**
	 * Returns list of dependencies which has given property as dependent property. Used locally.
	 * 
	 * @param property 	dependent property
	 * 
	 * @return 			dependencies which has given property as dependent property
	 */
	private List<Dependency> getOutputDependencies(Property property)
	{
		ArrayList<Dependency> outputDependencies = new ArrayList<Dependency>();
		
		for (Dependency dependency : getAllDependenciesList())
		{
			
			if (dependency.getIndependent() == property)
			{
				outputDependencies.add(dependency);
			}
		}
		
		return outputDependencies;
	}
	
	
	/**
	 * Adds transform between given properties. Used locally.
	 * 
	 * @param sourceProperty 	transform source
	 * 
	 * @param destProperty		transform destiny
	 */
	private void addTransform(Property sourceProperty, Property destProperty)
	{
		Transform transform = new Transform(sourceProperty, destProperty);
	
		transformsList.add(transform);
	}
	
	
	/**
	 * Adds dependency between given properties. Used locally.
	 * 
	 * @param independentProperty	
	 * 
	 * @param dependentProperty
	 */
	private void addDependency(Property independentProperty, Property dependentProperty)
	{
		Dependency dependency = new Dependency(independentProperty, dependentProperty);
		
		dependenciesList.add(dependency);
	}
	
	
	/**
	 * Returns list of all dependencies that can exist between model properties.
	 * Model holds only dependencies between properties that exist in last phase of ARD design.
	 * But sometimes (i.e. if we use "Roll" option on ARD graph) we need information about dependencies
	 * between properties that are not normally visible on ARD diagram. Used locally. 
	 * 
	 * @return	list of all possible dependencies
	 */
	private List<Dependency> getAllDependenciesList()
	{
		List<Dependency> newDependenciesList = new ArrayList<Dependency>();
		
		// all possible dependencies contains dependencies from model
		newDependenciesList.addAll(dependenciesList);
		
		boolean addDependency = true;
		
		while (addDependency)
		{
			addDependency = false;
			
			// check each one property
			for (Property property : propertiesList)
			{
				List<Property> transformDestinies = this.getTransformDestinies(property);
				
				// if property is source property of finalization
				if (transformDestinies.size() == 1)
				{
					// get destiny for that finalization
					Property transformDestiny = transformDestinies.get(0);
					
					List<Dependency> newDependencies = new ArrayList<Dependency>();
					
					// if returned list contain dependencies with considered property as independent or dependent property,
					// add dependency with the same independent/dependent property, but witch dependent/independent property 
					// matching the finalization destiny.
					// In other words, if ARD graph contain some dependency A->B, then it can also contain dependency
					// with the same independent property A, but with dependent property C created as a result of
					// finalization of B (A->C), or with the same dependent property B, but with independent property C
					// created as a result of finalization of A (C->B)
					for (Dependency dependency : newDependenciesList)
					{
						if (dependency.getIndependent() == transformDestiny)
						{
							Dependency newDependency = new Dependency(property, dependency.getDependent());
								
							if(!dependencyExist(newDependenciesList, newDependency))
							{
								newDependencies.add(newDependency);
								addDependency = true;
							}
						}
						
						if (dependency.getDependent() == transformDestiny)
						{
							Dependency newDependency = new Dependency(dependency.getIndependent(), property);
								
							if(!dependencyExist(newDependenciesList, newDependency))
							{
								newDependencies.add(newDependency);
								addDependency = true;
							}
						}
					}
					
					newDependenciesList.addAll(newDependencies);
				}
				
				// if property is base property of split operation
				if (transformDestinies.size() > 1)
				{
					// get list of child properties for considered property
					List<Property> childProperties = new ArrayList<Property>();
					
					recurrentSetChildProperties(childProperties, property);
					
					List<Dependency> newDependencies = new ArrayList<Dependency>();
					
					// if dependency A->B exists in returned list, then returned list should also contain
					// dependencies between child properties of property A and child properties of property B
					for (Dependency dependency : newDependenciesList)
					{
						if (childProperties.contains(dependency.getIndependent()) && !childProperties.contains(dependency.getDependent()))
						{
							Dependency newDependency = new Dependency(property, dependency.getDependent());
								
							if(!dependencyExist(newDependenciesList, newDependency))
							{
								newDependencies.add(newDependency);
								addDependency = true;
							}
						}
						
						if (childProperties.contains(dependency.getDependent()) && !childProperties.contains(dependency.getIndependent()))
						{
							Dependency newDependency = new Dependency(dependency.getIndependent(), property);
								
							if(!dependencyExist(newDependenciesList, newDependency))
							{
								newDependencies.add(newDependency);
								addDependency = true;
							}
						}
					}
					
					newDependenciesList.addAll(newDependencies);
				}
			}
		}
		
		List<Dependency> removedDependencies = new ArrayList<Dependency>();
		
		for (Dependency dependency : newDependenciesList)
		{
			if ((dependency.getIndependent() == dependency.getDependent() && this.getTransformDestinies(dependency.getIndependent()).size() > 0) ||
					this.getTransformDestinies(dependency.getIndependent()).contains(dependency.getDependent()) ||
					this.getTransformDestinies(dependency.getDependent()).contains(dependency.getIndependent()))
			{
				removedDependencies.add(dependency);
			}
		}
		
		newDependenciesList.removeAll(removedDependencies);
		
		return newDependenciesList;
	}
	
	
	/**
	 * Find child properties of given property. Child properties are properties that lays under given property
	 * on TPH graph. Used locally.
	 * 
	 * @param childProperties 	list of child properties for given property
	 * 
	 * @param property			actually considered property
	 */
	private void recurrentSetChildProperties(List<Property> childProperties, Property property)
	{
		// for each transform destinies to given property
		for (Property prop : this.getTransformDestinies(property))
		{
			// add destiny property to list
			childProperties.add(prop);
			
			// go one level down
			recurrentSetChildProperties(childProperties, prop);
		}
	}
	
	
	/**
	 * Removes not necessary dependencies from model, which could appear i.e. as a result of finalization or split operation. 
	 */
	private void clearDependencies()
	{
		List<Dependency> removedDependencies = new ArrayList<Dependency>();
		
        for (Dependency dependency : getDependenciesList())
        {
        	Property independentProperty = dependency.getIndependent();
        	
        	Property dependentProperty = dependency.getDependent();
        	
        	// if this is not dependency exist in last phase of ARD design, remove it.
        	// In other words, if independent property or dependent property of given dependency is not a leaf on TPH graph,
        	// it means we can have more detailed ARD graph
        	if (this.getTransformDestinies(independentProperty).size() != 0 || this.getTransformDestinies(dependentProperty).size() != 0)
        	{
        		removedDependencies.add(dependency);
        	}
     
        	// if independent property or dependent property of given dependency doesn't exist, remove that dependency
        	if (!this.getPropertiesList().contains(dependency.getIndependent()) ||
        			!this.getPropertiesList().contains(dependency.getDependent()))
        	{
        		removedDependencies.add(dependency);
        	}
        }
        
        getDependenciesList().removeAll(removedDependencies);
	}
	
	
	private void clearTransforms()
	{
		List<Transform> transformsList = getTransformsList();
		List<Property> propertiesList = getPropertiesList();
		
		if (transformsList != null && propertiesList != null)
		{
			List<Transform> removedTransforms = new ArrayList<Transform>();
			List<Property> removedProperties = new ArrayList<Property>();
			
	        for (Transform transform : transformsList)
	        {
	        	if (!propertiesList.contains(transform.getSource()) || ! propertiesList.contains(transform.getDestiny()))
	        	{
	        		removedTransforms.add(transform);
	        	}
	        	
	        	if (transform.getSource().getAttributes().containsAll(transform.getDestiny().getAttributes()) &&
	        			transform.getDestiny().getAttributes().containsAll(transform.getSource().getAttributes()))
	        	{
	        		removedTransforms.add(transform);
	        		removedProperties.add(transform.getDestiny());
	        		
	        		List<Property> destinyProperties = getTransformDestinies(transform.getDestiny());
	        		
	        		for (Transform tempTransform : getTransformsList())
	        		{
	        			if (tempTransform.getSource() == transform.getDestiny() && destinyProperties.contains(tempTransform.getDestiny()))
	        			{
	        				tempTransform.setSource(transform.getSource());
	        			}
	        		}
	        	}
	        }
	        
	        propertiesList.removeAll(removedProperties);
	        transformsList.removeAll(removedTransforms);
		}
	}
	
	
	/**
	 * Removes all attributes, properties, destinies and transforms which were created from given property.
	 * It means this method removes from graphs everything laying down for given property. As a result given
	 * property exists as a last phase of ARD design. Used locally.
	 * 
	 * @param property	property from which start to remove
	 * @param removeProperty	true if given property have to be removed
	 */
	private void removeDownFromProperty(Property property, boolean removeProperty)
	{
		List<Attribute> removedAttributes = new ArrayList<Attribute>();
		List<Property> removedProperties = new ArrayList<Property>();
		List<Dependency> removedDependencies = new ArrayList<Dependency>();
		List<Transform> removedTransforms = new ArrayList<Transform>();
		
		List<Property> transformDestinies = getTransformDestinies(property);
		
		// create list of properties that lays under given property on TPH graph
		for (Property destinyProperty : transformDestinies)
		{
			recurentRemoveDestiniesProperties(removedProperties, destinyProperty);
		}
		
		if (removeProperty) removedProperties.add(property);
		
		propertiesList.removeAll(removedProperties);
		
		
		// remove dependencies between properties which has been removed
		for (Dependency dependency : getDependenciesList())
		{
			if (removedProperties.contains(dependency.getIndependent()) ||
					removedProperties.contains(dependency.getDependent()))
			{
				removedDependencies.add(dependency);
			}
		}

		getDependenciesList().removeAll(removedDependencies);
		
		
		// remove all transforms that has source or destiny property removed
		for (Transform transform : getTransformsList())
		{
			if (removedProperties.contains(transform.getSource()) ||
					removedProperties.contains(transform.getDestiny()))
			{
				removedTransforms.add(transform);
			}
		}
		
		getTransformsList().removeAll(removedTransforms);
		
		
		// remove all attributes that are unused
		for (Attribute attr : getAttributesList())
		{
			removedAttributes.add(attr);
		}

		List<Property> propertiesList = getPropertiesList();

		for (Property prop : propertiesList)
		{
			for (Attribute attr : prop.getAttributes())
			{
				removedAttributes.remove(attr);
			}
		}
		
		attributesList.removeAll(removedAttributes);
	}
	
	
	/**
	 * Find all properties that lays under given property on TPH graph. Used locally.
	 * 
	 * @param removedProperties		list of found properties
	 * 
	 * @param property				actually considered property
	 */
	private void recurentRemoveDestiniesProperties(List<Property> removedProperties, Property property)
	{
		List<Property> transformDestinies = getTransformDestinies(property);
		
		for (Property destinyProperty : transformDestinies)
		{
			recurentRemoveDestiniesProperties(removedProperties, destinyProperty);
		}
		
		removedProperties.add(property);
	}
	
	
	/**
	 * Adds attribute to properties. Used locally.
	 * 
	 * @param property		actually considered property
	 * 
	 * @param newAttribute	added attribute
	 */
	private void recurentAddAttributeToProperty(Property property, Attribute newAttribute)
	{
		List<Property> sourceProperties = this.getTransformSources(property);
		
		List<Property> destinyProperties = this.getTransformDestinies(property);
		
		// new attribute can appear only as a result of finalization.
		// So continue processing until reach finalization operation.
		if (destinyProperties.size() != 1)
		{
			List<Attribute> attributesList = property.getAttributes();
			
			attributesList.add(newAttribute);
			
			if (sourceProperties.size() > 0)
			{
				recurentAddAttributeToProperty(sourceProperties.get(0), newAttribute);
			}
		}
	}
	
	
	/**
	 * Checks if given dependency is the same as one of dependencies in list.
	 * 
	 * @param dependenciesList
	 * 
	 * @param dependency
	 * 
	 * @return	true if dependency exists in given list, false otherwise.
	 */
	private boolean dependencyExist(List<Dependency> dependenciesList, Dependency dependency)
	{
		for (Dependency dep : dependenciesList)
		{
			if (dep.getIndependent() == dependency.getIndependent() &&
					dep.getDependent() == dependency.getDependent())
			{
				return true;
			}
		}
		
		return false;
	}
}
