/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import java.awt.event.*;
import java.awt.*;
import javax.swing.*;
import org.jgraph.JGraph;
import pl.edu.agh.hekate.ard.actions.*;


/**
 * Mouse event handler for ARD graph.
 * 
 * @author Szymon Książek
 *
 */
public class ARDMouseListener extends MouseAdapter 
{
	/**
	 * Model instance
	 */
	private ARDModel model = null;
	
	/**
	 * View instance
	 */
	private ARDView view = null;
	
	/**
	 * Specifies last coordinates of dragged point
	 */
	private Point dragStartPoint = null;
	
	
	/**
	 * Constructor
	 * 
	 * @param model model instance
	 * @param view  view instance
	 */
	public ARDMouseListener(ARDModel model, ARDView view)
	{
		this.model = model;
		this.view = view;
	}
	
	/**
	 * Invoked when a mouse button has been double pressed on ARD graph 
	 */
	public void mousePressed(MouseEvent e)
	{
		if (!e.isControlDown() && e.getButton() == MouseEvent.BUTTON1 && e.getClickCount() == 2)
		{	
				GraphNode selectedGraphNode = null;
				
				Component component = e.getComponent();
				
				JGraph ARDGraph = (JGraph)component;
			
				// get selected cell
				Object cell = ARDGraph.getFirstCellForLocation(e.getX(), e.getY());
				
				if (cell != null && cell instanceof GraphNode)
				{
					selectedGraphNode = (GraphNode)cell;
					
					view.markAllDown(selectedGraphNode.getProperty());
				}
				else
				{
					view.markAllDown(null);
				}
		}
		else if (e.isControlDown())
		{
			GraphNode selectedGraphNode = null;
			
			Component component = e.getComponent();
			
			JGraph ARDGraph = (JGraph)component;
		
			// get selected cell
			Object cell = ARDGraph.getFirstCellForLocation(e.getX(), e.getY());
			
			if (cell != null && cell instanceof GraphNode)
			{
				selectedGraphNode = (GraphNode)cell;
				
				view.selectProperty(selectedGraphNode.getProperty());
			}
			else
			{
				view.selectProperty(null);
			}
		}
		else if (e.isAltDown())
		{	
			dragStartPoint = e.getLocationOnScreen();
		}
		
		if (e.getButton() == MouseEvent.BUTTON3)
		{
			boolean showAddRootNodeItem = false;
			boolean showAddFinalizationItem = false;
			boolean showAddSplitItem = false;
			boolean showRollItem = false;
			boolean showUnrollItem = false;
			boolean showUnrollAllItem = false;
			
			
			GraphNode selectedGraphNode = null;
			
			// if ARD graph is empty, only "Add root" item is allowed
			if (view.getARDGraphNodes().size() == 0)
			{
				showAddRootNodeItem = true;
			}
			else
			{
				Component component = e.getComponent();
			
				JGraph ARDGraph = (JGraph)component;
			
				Object cell = ARDGraph.getFirstCellForLocation(e.getX(), e.getY());
				
				if (cell != null && cell instanceof GraphNode)
				{
					selectedGraphNode = (GraphNode)cell;
					
					view.markAllDown(selectedGraphNode.getProperty());
					
					// if selected property contains for one conceptual attribute
					if (selectedGraphNode.getProperty().getAttributes().size() == 1 &&
							selectedGraphNode.getProperty().getAttributes().get(0).getName().charAt(0) ==
							selectedGraphNode.getProperty().getAttributes().get(0).getName().toUpperCase().charAt(0))
					{
						showAddFinalizationItem = true;
					}
					
					// if selected property contains from more than one attribute
					if (selectedGraphNode.getProperty().getAttributes().size() > 1)
					{
						showAddSplitItem = true;
					}
					
					// if selected property is not a root node on TPH graph
					if (model.getTransformSources(selectedGraphNode.getProperty()).size() > 0)
					{
						showRollItem = true;
					}
					
					// if selected property does not contain to the last design phase
					if (model.getTransformDestinies(selectedGraphNode.getProperty()).size() > 0)
					{
						showUnrollItem = true;
					}
				}
				else
				{
					view.markAllDown(null);
				}
			}
			
			if (model.getPropertiesList().size() > 0)
			{
				showUnrollAllItem = true;
			}
			
			
			// create and show context menu
			JPopupMenu popupMenu = new JPopupMenu();
			
			if (showAddRootNodeItem)
			{
				JMenuItem addRootNodeItem = new JMenuItem("Add root");
				addRootNodeItem.setMnemonic('r');
				addRootNodeItem.addActionListener(new AddRootNodeAction(model, view));
				addRootNodeItem.setIcon(view.getArranger().getIcon("add_root"));
				popupMenu.add(addRootNodeItem);
			}
			
			if (showAddFinalizationItem)
			{
				JMenuItem addFinalizationItem = new JMenuItem("Finalization");
				addFinalizationItem.setMnemonic('F');
				addFinalizationItem.addActionListener(new AddFinalizationAction(model, view, selectedGraphNode));
				addFinalizationItem.setIcon(view.getArranger().getIcon("finalization"));
				popupMenu.add(addFinalizationItem);
			}
			
			if (showAddSplitItem)
			{
				JMenuItem addSplitItem = new JMenuItem("Split");
				addSplitItem.setMnemonic('S');
				addSplitItem.addActionListener(new AddSplitAction(model, view, selectedGraphNode));
				addSplitItem.setIcon(view.getArranger().getIcon("split"));
				popupMenu.add(addSplitItem);
			}
			
			if (showRollItem)
			{
				JMenuItem rollItem = new JMenuItem("Roll");
				rollItem.setMnemonic('R');
				rollItem.addActionListener(new RollAction(view, selectedGraphNode, true));
				rollItem.setIcon(view.getArranger().getIcon("roll"));
				popupMenu.add(rollItem);
			}
			
			if (showUnrollItem)
			{
				JMenuItem unrollItem = new JMenuItem("Unroll");
				unrollItem.setMnemonic('U');
				unrollItem.addActionListener(new RollAction(view, selectedGraphNode, false));
				unrollItem.setIcon(view.getArranger().getIcon("unroll"));
				popupMenu.add(unrollItem);
			}
			
			if (showUnrollAllItem)
			{
				JMenuItem unrollAllItem = new JMenuItem("Unroll all");
				unrollAllItem.setMnemonic('a');
				
				unrollAllItem.addActionListener(new
						ActionListener()
				{
					public void actionPerformed(ActionEvent e)
					{
						view.unrollAll();
					}
				});
				
				unrollAllItem.setIcon(view.getArranger().getIcon("unroll_all"));
				popupMenu.add(unrollAllItem);
			}
			
			
			if (popupMenu.getPopupMenuListeners().length > 0)
			{
				popupMenu.show(e.getComponent(), e.getX(), e.getY());
			}
		}
	}
	
	// invoked when TPH graph is dragged
	public void mouseDragged(MouseEvent e)
	{
		if (e.isAltDown())
		{
			Point actualPoint = e.getLocationOnScreen();
			
			int xMove = (int)(actualPoint.getX() - dragStartPoint.getX());
			int yMove = (int)(actualPoint.getY() - dragStartPoint.getY());
			
			if (xMove != 0 || yMove != 0)
			{
				view.dragADRGraph(xMove, yMove);
			}
			
			dragStartPoint = actualPoint;
		}
	}
	
	
	public void mouseWheelMoved(MouseWheelEvent e)
	{
		// number of 'clicks' - how fast wheel is rotated
		int clicks = e.getWheelRotation();
		
		// zoom out ARD graph
		if (clicks > 0)
		{
			for (int zoomCount = 0 ; zoomCount < clicks ; zoomCount++)
			{
				view.zoomARDGraph(false);
			}
		}
		// zoom out ARD graph
		else if (clicks < 0)
		{
			clicks = -clicks;
			
			for (int zoomCount = 0 ; zoomCount < clicks ; zoomCount++)
			{
				view.zoomARDGraph(true);
			}
		}
	}
}