/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import java.util.*;
import javax.swing.*;
import org.jgraph.JGraph;

/**
 * View interface
 * 
 * @author Szymon Książek
 *
 */
public interface ARDView 
{
	/**
	 * Creates new ARD graph
	 */
	public void createARDGraph();
	
	
	/**
	 * Repaints ARD graph
	 */
	public void repaintARDGraph();
	
	
	/**
	 * Refreshes (create and repaint) ARD graph
	 */
	public void refreshARDGraph();
	
	
	/**
	 * Rotates ARD graph
	 * 
	 * @param rotateRight 	true for rotate right, false for rotate left
	 */
	public void rotateARDGraph(boolean rotateRight);
	
	
	/**
	 * Adds new node to ARD graph
	 * 
	 * @param graphNode		added graph node
	 */
	public void addARDGraphNode(GraphNode graphNode);
	
	
	/**
	 * Removes node from ARD graph
	 * 
	 * @param graphNode		removed graph node
	 */
	public void removeARDGraphNode(GraphNode graphNode);
	
	
	/**
	 * Sets ARD graph layout
	 * 
	 * @param layoutType	new ARD graph layout
	 */
	public void setARDLayout(LayoutType layoutType);
	
	
	/**
	 * Creates new TPH graph
	 */
	public void createTPHGraph();
	
	
	/**
	 * Repaints TPH graph
	 */
	public void repaintTPHGraph();
	
	
	/**
	 * Refreshes (create and repaint) TPH graph
	 */
	public void refreshTPHGraph();
	
	
	/**
	 * Rotates TPH graph
	 * 
	 * @param rotateRight 	true for rotate right, false for rotate left
	 */
	public void rotateTPHGraph(boolean rotateRight);
	
	
	/**
	 * Sets TPH graph layout
	 * 
	 * @param layoutType	new ARD graph layout
	 */
	public void setTPHLayout(LayoutType layoutType);
	
	
	/**
	 * Refreshes both ARD and TPH graphs
	 */
	public void refreshGraphs();
	
	
	/**
	 * Changes zoom of ARD graph
	 * 
	 * @param zoomIn	true to zoom in, false to zoom out
	 */
	public void zoomARDGraph(boolean zoomIn);
	
	
	/**
	 * Changes zoom of TPH graph
	 * 
	 * @param zoomIn	true to zoom in, false to zoom out
	 */
	public void zoomTPHGraph(boolean zoomIn);
	
	
	/**
	 * Changes distance between nodes on ARD graph
	 * 
	 * @param increaseLevelDistance	true to increase distance between nodes, false to decrease distance
	 */
	public void changeARDNodesDistance(boolean increaseLevelDistance, boolean horizontal);
	
	
	/**
	 * Changes distance between nodes on TPH graph
	 * 
	 * @param increaseLevelDistance	true to increase distance between nodes, false to decrease distance
	 */
	public void changeTPHNodesDistance(boolean increaseLevelDistance, boolean horizontal);
	

	/**
	 * Returns panel that contains ARD graph
	 * 
	 * @return	ARD panel
	 */
	public JPanel getARDPanel();
	
	
	/**
	 * Returns panel that contains TPH graph
	 * 
	 * @return	TPH panel
	 */
	public JPanel getTPHPanel();
	
	
	/**
	 * Returns list of all ARD graph nodes
	 * 
	 * @return 	list of nodes from ARD graph
	 */
	public ArrayList<GraphNode> getARDGraphNodes();
	
	
	/**
	 * Sets ARD graph nodes
	 * 
	 * @param graphNodes	list of ARD graph nodes
	 */
	public void setARDGraphNodes(ArrayList<GraphNode> graphNodes);
	
	
	/**
	 * Returns list of all TPH graph nodes
	 * 
	 * @return 	list of nodes from TPH graph
	 */
	public ArrayList<GraphNode> getTPHGraphNodes();
	
	
	/**
	 * Returns instance of persistence manager
	 * 
	 * @return	persistence manager instance
	 */
	public PersistenceManager getPersistenceManager();
	
	
	/**
	 * Sets scale of ARD graph
	 * 
	 * @param scale		ARD graph scale
	 */
	public void setARDScale(double scale);
	
	
	/**
	 * Sets scale of TPH graph
	 * 
	 * @param scale		TPH graph scale
	 */
	public void setTPHScale(double scale);
	

	/**
	 * Moves ARD graph
	 * 
	 * @param x		horizontal distance
	 * @param y		vertical distance
	 */
	public void dragADRGraph(int x, int y);
	
	
	/**
	 * Moves TPH graph
	 * 
	 * @param x		horizontal distance
	 * @param y		vertical distance
	 */
	public void dragTPHGraph(int x, int y);
	
	
	/**
	 * Marks all nodes down starting from given property
	 * 
	 * @param property		initial property
	 */
	public void markAllDown(Property property);
	
	
	/**
	 * Rolls given ARD graph. Used when user select "Roll" option from menu.
	 * 
	 * @param graphNode		ARD graph node to roll
	 */
	public void rollItem(GraphNode graphNode);
	
	
	/**
	 * Unolls given ARD graph node Used when user select "Unroll" option from menu.
	 * 
	 * @param graphNode		ARD graph node to unroll
	 */
	public void unrollItem(GraphNode graphNode);
	
	
	/**
	 * Unrolls all ARD graph nodes. Used when user select "Unroll all" option from menu.
	 */
	public void unrollAll();
	
	
	/**
	 * Deletes from TPH graph all nodes that are not associated with existing properties
	 * and create most detailed ARD graph
	 */
	public void clearGraphs();
	
	
	/**
	 * Returns instance of arranger implementation.
	 * 
	 * @return	arranger instance
	 */
	public Arranger getArranger();
	
	
	/**
	 * Returns current ARD graph instance
	 * 
	 * @return	ARD graph instance
	 */
	public JGraph getARDGraph();
	
	
	/**
	 * Sets new ARD graph instance
	 * 
	 * @param graph		new ARD graph instance
	 */
	public void setARDGraph(JGraph graph);
	
	
	/**
	 * Returns current TPH graph instance
	 * 
	 * @return	TPH graph instance
	 */
	public JGraph getTPHGraph();
	
	
	/**
	 * Sets new TPH graph instance
	 * 
	 * @param graph		new TPH graph instance
	 */
	public void setTPHGraph(JGraph graph);
	
	
	/**
	 * Used to export ARD graph to image
	 */
	public void exportARDToImage();
	
	
	/**
	 * Used to export TPH graph to image
	 */
	public void exportTPHToImage();
	
	
	/**
	 * Used to export both ARD and TPH graphs to images
	 */
	public void exportBothToImage();
	
	
	/**
	 * Used to export graph to Prolog
	 */
	public void exportToProlog();
	
	
	/**
	 * Used to select property on ARD graph
	 * 
	 * @param property	selected property
	 */
	public void selectProperty(Property property);
	
	/**
	 * Optimizes ARD graph size
	 */
	public void optimizeARDSize();
	
	/**
	 * Optimizes TPH graph size
	 */
	public void optimizeTPHSize();
}
