/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import javax.imageio.ImageIO;
import javax.swing.*;
import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.*;
import java.util.List;		// bo konflikt z java.awt.List

import org.jgraph.JGraph;
import org.jgraph.graph.CellView;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.EdgeView;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.GraphModel;
import org.jgraph.graph.DefaultPort;

import pl.edu.agh.hekate.ard.actions.*;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.JGraphLayout;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;
import com.jgraph.layout.organic.JGraphFastOrganicLayout;
import com.jgraph.layout.organic.JGraphSelfOrganizingOrganicLayout;
import com.jgraph.layout.orthogonal.OrthogonalLinkRouter;
import com.jgraph.layout.routing.JGraphParallelRouter;
import com.jgraph.layout.tree.JGraphTreeLayout;


/**
 * Implemetation of view
 *  
 * @author Szymon Książek
 */
public class ARDViewImpl implements ARDView
{
	/**
	 * Model instance
	 */
	private ARDModel model = null; 
	
	/**
	 * Persistence manager instance
	 */
	private PersistenceManager persistenceManager = null;
	
	/**
	 * Arranger instance
	 */
	private Arranger arranger = null;
	
	/**
	 * Panel that contains ARD graph
	 */
	private JPanel ardPanel;
	
	/**
	 * Panel that contains TPH graph
	 */
	private JPanel tphPanel; 

	/**
	 * Scroll pane that contains ARD panel
	 */
	private JScrollPane ardScrollPane;
	
	/**
	 * Scroll pane that contains TPH panel
	 */
	private JScrollPane tphScrollPane; 
	
	/**
	 * Model of ARD graph
	 */
	private GraphModel ardGraphModel;
	
	/**
	 * Model of TPH graph
	 */
	private GraphModel tphGraphModel; 
	
	/**
	 * Instance of ARD graph
	 */
	private JGraph ardGraph; 
	
	/**
	 * Instance of TPH graph
	 */
	private JGraph tphGraph; 
	
	/**
	 * Nodes from ARD graph
	 */
	private ArrayList<GraphNode> ardGraphNodes;
	
	/**
	 * Nodes from TPH graph
	 */
	private ArrayList<GraphNode> tphGraphNodes; 
	
	/**
	 * Indicates scale of ARD graph
	 */
	private double ardScale; 
	
	/**
	 * Indicates scale of TPH graph
	 */
	private double tphScale; 
	
	/**
	 * Cells from ARD graph (both nodes and edges)
	 */
	private DefaultGraphCell[] ardCells = null; 
	
	/**
	 * Cells from TPH graph (both nodes and edges)
	 */
	private DefaultGraphCell[] tphCells = null; 
	
	/**
	 * Indicates orientation of ARD graph
	 */
	private int ardOrientation;						 
	
	/**
	 * Indicates layout type of ARD graph
	 */
	private LayoutType ardLayout;					 
	
	/**
	 * Indicates orientation of TPH graph
	 */
	private int tphOrientation;						 
	
	/**
	 * Indicates layout type of TPH graph
	 */
	private LayoutType tphLayout;	
	
	/**
	 * Property that was selected with CTRL button 
	 */
	private Property selectedProperty = null;
	
	/**
	 * Main frame reference
	 */
	private JFrame mainFrame = null;
	
	
	/**
	 * Constructor.
	 * 
	 * @param model					model instance
	 * 
	 * @param persistenceManager	persistence manager instance
	 */
	public ARDViewImpl(ARDModel model, PersistenceManager persistenceManager, JFrame frame)
	{
		this.model = model;
		this.persistenceManager = persistenceManager;
		this.mainFrame = frame;
		
		ardGraphNodes = new ArrayList<GraphNode>();
		tphGraphNodes = new ArrayList<GraphNode>();
		
		ardPanel = new JPanel(new BorderLayout());
		tphPanel = new JPanel(new BorderLayout());
		
		arranger = new ArrangerImpl();
		
		ardLayout = LayoutType.TREE_LAYOUT;
		ardOrientation = SwingConstants.WEST;
		
		tphLayout = LayoutType.TREE_LAYOUT;
		tphOrientation = SwingConstants.NORTH;
		
		ardScale = 1.0;
		tphScale = 1.0;
		
		ardCells = new DefaultGraphCell[0];
		tphCells = new DefaultGraphCell[0];
		
		
		// create toolbar for ARD tools
		JPanel ARDToolBarsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));

		JToolBar ARDLayoutTypeToolBar = new JToolBar();
		
		Action setARDTreeLayoutAction = new LayoutTypeChangeAction("Tree layout", arranger.getIcon("tree_layout"), "ARD", LayoutType.TREE_LAYOUT, ARDViewImpl.this);
		Action setARDOrganicLayoutAction = new LayoutTypeChangeAction("Organic layout", arranger.getIcon("organic_layout"), "ARD", LayoutType.ORGANIC_LAYOUT, ARDViewImpl.this);
		
		ARDLayoutTypeToolBar.add(setARDTreeLayoutAction);
		ARDLayoutTypeToolBar.add(setARDOrganicLayoutAction);
		
		ARDLayoutTypeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		ARDLayoutTypeToolBar.setName("ARD layout");
		
		ARDToolBarsPanel.add(ARDLayoutTypeToolBar);
		
		
		JToolBar ARDRotateGraphToolBar = new JToolBar();
		
		Action rotateRightARDGraphAction = new RotateGraphAction("Rotate right", arranger.getIcon("rotate_right"), "ARD", true, ARDViewImpl.this);
		Action rotateLeftARDGraphAction = new RotateGraphAction("Rotate left", arranger.getIcon("rotate_left"), "ARD", false, ARDViewImpl.this);
		
		ARDRotateGraphToolBar.add(rotateLeftARDGraphAction);
		ARDRotateGraphToolBar.add(rotateRightARDGraphAction);
		
		ARDRotateGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		ARDRotateGraphToolBar.setName("Rotate ARD");
		
		ARDToolBarsPanel.add(ARDRotateGraphToolBar);
		
		
		JToolBar ARDZoomGraphToolBar = new JToolBar();
		
		Action zoomInARDGraphAction = new ZoomGraphAction("Zoom in", arranger.getIcon("zoom_in"), "ARD", true, ARDViewImpl.this);
		Action zoomOutARDGraphAction = new ZoomGraphAction("Zoom out", arranger.getIcon("zoom_out"), "ARD", false, ARDViewImpl.this);
		
		ARDZoomGraphToolBar.add(zoomInARDGraphAction);
		ARDZoomGraphToolBar.add(zoomOutARDGraphAction);
		
		ARDZoomGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		ARDZoomGraphToolBar.setName("Zoom ARD");
		
		ARDToolBarsPanel.add(ARDZoomGraphToolBar);
		
		
		JToolBar ARDChangeDistanceToolBar = new JToolBar();
		
		Action increaseARDHorizontalDistanceAction = new ChangeNodesDistanceAction("Increase level distance", arranger.getIcon("increase_horizontal_distance"), "ARD", "horizontal", true, ARDViewImpl.this);
		Action decreaseARDHorizontalDistanceAction = new ChangeNodesDistanceAction("Decrease level distance", arranger.getIcon("decrease_horizontal_distance"), "ARD", "horizontal", false, ARDViewImpl.this);
		Action increaseARDVerticalDistanceAction = new ChangeNodesDistanceAction("Increase nodes distance", arranger.getIcon("increase_vertical_distance"), "ARD", "vertical", true, ARDViewImpl.this);
		Action decreaseARDVerticalDistanceAction = new ChangeNodesDistanceAction("Decrease nodes distance", arranger.getIcon("decrease_vertical_distance"), "ARD", "vertical", false, ARDViewImpl.this);
		
		ARDChangeDistanceToolBar.add(increaseARDHorizontalDistanceAction);
		ARDChangeDistanceToolBar.add(decreaseARDHorizontalDistanceAction);
		ARDChangeDistanceToolBar.add(increaseARDVerticalDistanceAction);
		ARDChangeDistanceToolBar.add(decreaseARDVerticalDistanceAction);
		
		ARDChangeDistanceToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		ARDChangeDistanceToolBar.setName("Change ARD level distance");
		
		ARDToolBarsPanel.add(ARDChangeDistanceToolBar);
		
		
		JToolBar ARDOptimizeDiagramSizeToolBar = new JToolBar();
		
		Action optimizeARDSizeAction = new OptimizeDiagramSizeAction("Optimize ARD diagram size", arranger.getIcon("optimize_graph_size"), "ARD", ARDViewImpl.this);
		
		ARDOptimizeDiagramSizeToolBar.add(optimizeARDSizeAction);
		
		ARDOptimizeDiagramSizeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		ARDOptimizeDiagramSizeToolBar.setName("Optimize ARD diagram size");
		
		ARDToolBarsPanel.add(ARDOptimizeDiagramSizeToolBar);
		
		


		ardPanel.add(ARDToolBarsPanel, BorderLayout.NORTH);
		
		
		// create toolbar for TPH tools
		JPanel TPHToolBarsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		
		JToolBar TPHLayoutTypeToolBar = new JToolBar();
		
		Action setTPHTreeLayoutAction = new LayoutTypeChangeAction("Tree layout", arranger.getIcon("tree_layout"), "TPH", LayoutType.TREE_LAYOUT, ARDViewImpl.this);
		Action setTPHOrganicLayoutAction = new LayoutTypeChangeAction("Organic layout", arranger.getIcon("organic_layout"), "TPH", LayoutType.ORGANIC_LAYOUT, ARDViewImpl.this);
		
		TPHLayoutTypeToolBar.add(setTPHTreeLayoutAction);
		TPHLayoutTypeToolBar.add(setTPHOrganicLayoutAction);
		
		TPHLayoutTypeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		TPHLayoutTypeToolBar.setName("TPH layout");
		
		TPHToolBarsPanel.add(TPHLayoutTypeToolBar);
		
		
		JToolBar TPHRotateGraphToolBar = new JToolBar();
		
		Action rotateRightTPHGraphAction = new RotateGraphAction("Rotate right", arranger.getIcon("rotate_right"), "TPH", true, ARDViewImpl.this);
		Action rotateLeftTPHGraphAction = new RotateGraphAction("Rotate left", arranger.getIcon("rotate_left"), "TPH", false, ARDViewImpl.this);
		
		TPHRotateGraphToolBar.add(rotateLeftTPHGraphAction);
		TPHRotateGraphToolBar.add(rotateRightTPHGraphAction);
		
		TPHRotateGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		TPHRotateGraphToolBar.setName("Rotate TPH");
		
		TPHToolBarsPanel.add(TPHRotateGraphToolBar);
		
		
		JToolBar TPHZoomGraphToolBar = new JToolBar();
		
		Action zoomInTPHGraphAction = new ZoomGraphAction("Zoom in", arranger.getIcon("zoom_in"), "TPH", true, ARDViewImpl.this);
		Action zoomOutTPHGraphAction = new ZoomGraphAction("Zoom out", arranger.getIcon("zoom_out"), "TPH", false, ARDViewImpl.this);
		
		TPHZoomGraphToolBar.add(zoomInTPHGraphAction);
		TPHZoomGraphToolBar.add(zoomOutTPHGraphAction);
		
		TPHZoomGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		TPHZoomGraphToolBar.setName("Zoom TPH");
		
		TPHToolBarsPanel.add(TPHZoomGraphToolBar);
		
		
		JToolBar TPHChangeDistanceToolBar = new JToolBar();
		
		Action increaseTPHHorizontalDistanceAction = new ChangeNodesDistanceAction("Increase nodes distance", arranger.getIcon("increase_horizontal_distance"), "TPH", "horizontal", true, ARDViewImpl.this);
		Action decreaseTPHHorizontalDistanceAction = new ChangeNodesDistanceAction("Decrease nodes distance", arranger.getIcon("decrease_horizontal_distance"), "TPH", "horizontal", false, ARDViewImpl.this);
		Action increaseTPHVerticalDistanceAction = new ChangeNodesDistanceAction("Increase level distance", arranger.getIcon("increase_vertical_distance"), "TPH", "vertical", true, ARDViewImpl.this);
		Action decreaseTPHVerticalDistanceAction = new ChangeNodesDistanceAction("Decrease level distance", arranger.getIcon("decrease_vertical_distance"), "TPH", "vertical", false, ARDViewImpl.this);
		
		TPHChangeDistanceToolBar.add(increaseTPHHorizontalDistanceAction);
		TPHChangeDistanceToolBar.add(decreaseTPHHorizontalDistanceAction);
		TPHChangeDistanceToolBar.add(increaseTPHVerticalDistanceAction);
		TPHChangeDistanceToolBar.add(decreaseTPHVerticalDistanceAction);
		
		TPHChangeDistanceToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		TPHChangeDistanceToolBar.setName("Change TPH level distance");
		
		TPHToolBarsPanel.add(TPHChangeDistanceToolBar);
		
		
		JToolBar TPHOptimizeDiagramSizeToolBar = new JToolBar();
		
		Action optimizeTPHSizeAction = new OptimizeDiagramSizeAction("Optimize TPH diagram size", arranger.getIcon("optimize_graph_size"), "TPH", ARDViewImpl.this);
		
		TPHOptimizeDiagramSizeToolBar.add(optimizeTPHSizeAction);
		
		TPHOptimizeDiagramSizeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		TPHOptimizeDiagramSizeToolBar.setName("Optimize TPH diagram size");
		
		TPHToolBarsPanel.add(TPHOptimizeDiagramSizeToolBar);
		
		
		
		
		tphPanel.add(TPHToolBarsPanel, BorderLayout.NORTH);
		
		
		// create ARD and TPH scroll panes
		ardScrollPane = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		tphScrollPane = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		
		ardScrollPane.getHorizontalScrollBar().setUnitIncrement(60);
		ardScrollPane.getHorizontalScrollBar().setBlockIncrement(180);
		ardScrollPane.getVerticalScrollBar().setUnitIncrement(60);
		ardScrollPane.getVerticalScrollBar().setBlockIncrement(180);
		
		tphScrollPane.getHorizontalScrollBar().setUnitIncrement(60);
		tphScrollPane.getHorizontalScrollBar().setBlockIncrement(180);
		tphScrollPane.getVerticalScrollBar().setUnitIncrement(60);
		tphScrollPane.getVerticalScrollBar().setBlockIncrement(180);
		
		
		// create empty ARD and TPH graphs
		ardGraphModel = new DefaultGraphModel();
		tphGraphModel = new DefaultGraphModel();
		
		ardGraph = new JGraph(ardGraphModel);
		tphGraph = new JGraph(tphGraphModel);
		
		
		ARDMouseListener ardMouseListener = new ARDMouseListener(model, ARDViewImpl.this);
		
		ardGraph.addMouseListener(ardMouseListener);
		ardGraph.addMouseMotionListener(ardMouseListener);
		ardGraph.addMouseWheelListener(ardMouseListener);
		
		
		TPHMouseListener tphMouseListener = new TPHMouseListener(model, ARDViewImpl.this);
		
		tphGraph.addMouseListener(tphMouseListener);
		tphGraph.addMouseMotionListener(tphMouseListener);
		tphGraph.addMouseWheelListener(tphMouseListener);
		
		
		ardScrollPane = new JScrollPane(ardGraph);
		
		tphScrollPane = new JScrollPane(tphGraph);

		ardPanel.add(ardScrollPane, BorderLayout.CENTER);
		tphPanel.add(tphScrollPane, BorderLayout.CENTER);		
	}
	
	
	public void createARDGraph()
	{
		mainFrame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		
		ardGraphModel = new DefaultGraphModel();
		
		ardGraph = new JGraph(ardGraphModel);
		
		ardGraph.setScale(ardScale);
		
		
		for (int i = 0 ; i < ardGraphNodes.size() ; i++)
		{
			GraphNode graphNode = ardGraphNodes.remove(i);
			
			GraphNode newGraphNode = new GraphNode(graphNode.getProperty());
			
			ardGraphNodes.add(i, newGraphNode);
		}
		
		
		// list of both nodes and edges
		List<DefaultGraphCell> cellsList = new ArrayList<DefaultGraphCell>();
		
		// add all visible nodes to graph
		cellsList.addAll(ardGraphNodes);
		
		
		// add all edges to graph
		for (int i = 0 ; i < ardGraphNodes.size() ; i++)
		{
			for (int j = i ; j < ardGraphNodes.size() ; j++)
			{
				int dependencyOrder = 
					model.isDependencyAllowed(ardGraphNodes.get(i).getProperty(), ardGraphNodes.get(j).getProperty());
				
				if (dependencyOrder < 0)
				{
					DefaultEdge edge = new DefaultEdge();
					edge.setSource(cellsList.get(i).getChildAt(0));
					edge.setTarget(cellsList.get(j).getChildAt(0));
					GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_CLASSIC);
					GraphConstants.setEndFill(edge.getAttributes(), true);	
					GraphConstants.setRouting(edge.getAttributes(), new EdgeRouter());
					
					cellsList.add(edge);
				}
				
				if (dependencyOrder > 0)
				{
					DefaultEdge edge = new DefaultEdge();
					edge.setSource(cellsList.get(j).getChildAt(0));
					edge.setTarget(cellsList.get(i).getChildAt(0));
					GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_CLASSIC);
					GraphConstants.setEndFill(edge.getAttributes(), true);
					GraphConstants.setRouting(edge.getAttributes(), new EdgeRouter());
					
					cellsList.add(edge);
				}
			}
		}
		
		
		// change list to table
		ardCells = new DefaultGraphCell[cellsList.size()];
		
		for (int i = 0 ; i < cellsList.size() ; i++)
		{
			ardCells[i] = cellsList.get(i);
		}

		ardGraph.getGraphLayoutCache().insert(ardCells);

		JGraphFacade facade = new JGraphFacade(ardGraph);
		
		facade.setIgnoresUnconnectedCells(true);
		facade.setDirected(true);
		facade.setEdgePromotion(false);
		facade.setIgnoresUnconnectedCells(false);
		facade.setOrdered(true);
		
		
		// create graph layout
		JGraphLayout layout = null;
		
		if (ardLayout == LayoutType.TREE_LAYOUT)
		{
			JGraphHierarchicalLayout graphHierarchicalLayout = new JGraphHierarchicalLayout();
			
			graphHierarchicalLayout.setOrientation(ardOrientation);
			graphHierarchicalLayout.setFineTuning(true);
			graphHierarchicalLayout.setFixRoots(true);
			graphHierarchicalLayout.setInterRankCellSpacing(arranger.getARDHorizontalNodesDistance());
			graphHierarchicalLayout.setIntraCellSpacing(arranger.getARDVerticalNodesDistance());
			
			layout = graphHierarchicalLayout;
		}
	
		if (ardLayout == LayoutType.ORGANIC_LAYOUT)
		{
			JGraphFastOrganicLayout graphOrganicLayout = new JGraphFastOrganicLayout();
			
			graphOrganicLayout.setForceConstant(arranger.getARDVerticalNodesDistance());
			
			layout = graphOrganicLayout;
		}
		
		layout.run(facade);
		
		Map nested = facade.createNestedMap(true, true);
		
		ardGraph.getGraphLayoutCache().edit(nested);
	 
	
		// originally, graph is placed in upper left corner.
		// Move graph 50 pixels right and 50 pixels down
		cds.jlow.client.graph.GraphJ.move(ardGraph, ardCells, 50, 50);
		
		ardGraph.setGridEnabled(arranger.isShowGrid());
		ardGraph.setGridVisible(arranger.isShowGrid());
	
		ARDMouseListener ardMouseListener = new ARDMouseListener(model, ARDViewImpl.this);
	
		ardGraph.addMouseListener(ardMouseListener);
		ardGraph.addMouseMotionListener(ardMouseListener);
		ardGraph.addMouseWheelListener(ardMouseListener);
	
		mainFrame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}
	
	
	public void repaintARDGraph()
	{	
		mainFrame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		
		ardPanel.remove(ardScrollPane);
		
		// set ARD graph background color
		ardGraph.setBackground(arranger.getARDBackgroundColor());
		
		Map nested=new Hashtable();
		
		// map that specifies colors of inactive nodes
		Map inactiveMap = new Hashtable();

		GraphConstants.setInset(inactiveMap, 10);
		GraphConstants.setBorderColor(inactiveMap, arranger.getARDUnmarkedLineColor());
		GraphConstants.setAutoSize(inactiveMap, true);
		GraphConstants.setOpaque(inactiveMap, true);
		GraphConstants.setSelectable(inactiveMap, true);
		GraphConstants.setEditable(inactiveMap, false);
		GraphConstants.setGradientColor(inactiveMap, arranger.getARDUnmarkedBackgroundColor());
		
		if (arranger.isGradientARDCell())
		{
			GraphConstants.setBackground(inactiveMap, Color.WHITE);
		}
		else
		{
			GraphConstants.setBackground(inactiveMap, arranger.getARDUnmarkedBackgroundColor());
		}
		
		
		// map that specifies colors of active nodes	
		Map activeMap = new Hashtable();

		GraphConstants.setGradientColor(activeMap, arranger.getARDMarkedBackgroundColor());
		GraphConstants.setOpaque(activeMap, true);
		GraphConstants.setInset(activeMap, 10);
		GraphConstants.setBorderColor(activeMap, arranger.getARDMarkedLineColor());
		GraphConstants.setAutoSize(activeMap, true);
		GraphConstants.setSelectable(activeMap, true);
		GraphConstants.setEditable(activeMap, false);
		
		if (arranger.isGradientARDCell())
		{
			GraphConstants.setBackground(activeMap, Color.WHITE);
		}
		else
		{
			GraphConstants.setBackground(activeMap, arranger.getARDMarkedBackgroundColor());
		}
		
		
		// map that specifies colors of selected nodes
		Map selectedMap = new Hashtable();
		
		GraphConstants.setBackground(selectedMap, Color.GREEN);
		GraphConstants.setGradientColor(selectedMap, arranger.getARDUnmarkedBackgroundColor());
		GraphConstants.setOpaque(selectedMap, true);
		
		// change colors of graph nodes
		for (int i = 0 ; i < ardCells.length ; i++)
		{
			if (ardCells[i] instanceof GraphNode)
			{
				GraphNode graphNode = (GraphNode)ardCells[i];
				
				if (graphNode.isSelected())
				{
					nested.put(graphNode, activeMap);
				}
				else
				{
					nested.put(graphNode, inactiveMap);
				}
				
				if (graphNode.getProperty() == selectedProperty)
				{
					nested.put(graphNode, selectedMap);
				}
			}
		}
		
		
		// map that specifies colors of inactive edges
		inactiveMap = new Hashtable();

    	GraphConstants.setLineColor(inactiveMap, arranger.getARDUnmarkedLineColor());

    	
    	// map that specifies colors of active nodes
		activeMap = new Hashtable();

		GraphConstants.setLineColor(activeMap, arranger.getARDMarkedLineColor());
		
		
		// change colors of edges
		for (int i = 0 ; i < ardCells.length ; i++)
		{
			if (ardCells[i] instanceof DefaultEdge)
			{
				DefaultEdge edge = (DefaultEdge)ardCells[i];
				
				DefaultGraphCell sourceCell = (DefaultGraphCell)(((DefaultPort)edge.getSource()).getParent());
				DefaultGraphCell targetCell = (DefaultGraphCell)((DefaultPort)edge.getTarget()).getParent();
				
				if (sourceCell instanceof GraphNode && targetCell instanceof GraphNode)
				{
					// if both source and target cell of edge are selected, then edge also is selected
					if (((GraphNode)sourceCell).isSelected() && ((GraphNode)targetCell).isSelected())
					{
						nested.put(edge, activeMap);
					}
					else
					{
						nested.put(edge, inactiveMap);
					}
				}			
			}
		}
		
		
		ardGraph.setGridEnabled(arranger.isShowGrid());
		ardGraph.setGridVisible(arranger.isShowGrid());
		
		ardGraph.getGraphLayoutCache().edit(nested,null,null,null);
		
		ardGraph.repaint();
		
		ardScrollPane = new JScrollPane(ardGraph);
		
		ardPanel.add(ardScrollPane, BorderLayout.CENTER);
		
		ardPanel.validate();
		
		ardGraph.setMinimumSize(new Dimension(ardGraph.getWidth() + arranger.getARDHorizontalNodesDistance(), ardGraph.getHeight()));
		
		if (arranger.isOptimizeGraphsSizes())
		{
			optimizeARDSize();
		}
		
		mainFrame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}
	
	
	public void refreshARDGraph()
	{
		createARDGraph();
		repaintARDGraph();	
	}
	
	
	public void rotateARDGraph(boolean rotateRight)
	{
		if (rotateRight)
		{
			if (ardOrientation == SwingConstants.WEST) ardOrientation = SwingConstants.NORTH;
			else if (ardOrientation == SwingConstants.NORTH) ardOrientation = SwingConstants.EAST;
			else if (ardOrientation == SwingConstants.EAST) ardOrientation = SwingConstants.SOUTH;
			else if (ardOrientation == SwingConstants.SOUTH) ardOrientation = SwingConstants.WEST;
		}
		else
		{
			if (ardOrientation == SwingConstants.NORTH) ardOrientation = SwingConstants.WEST;
			else if (ardOrientation == SwingConstants.EAST) ardOrientation = SwingConstants.NORTH;
			else if (ardOrientation == SwingConstants.SOUTH) ardOrientation = SwingConstants.EAST;
			else if (ardOrientation == SwingConstants.WEST) ardOrientation = SwingConstants.SOUTH;
		}
	}
	
	
	public void addARDGraphNode(GraphNode graphNode)
	{
		ardGraphNodes.add(graphNode);
	}
	
	
	public void removeARDGraphNode(GraphNode graphNode)
	{
		ardGraphNodes.remove(graphNode);
	}
	
	
	public void setARDLayout(LayoutType layoutType)
	{
		this.ardLayout = layoutType;
	}
	
	
	public void createTPHGraph()
	{
		mainFrame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		
		tphGraphModel = new DefaultGraphModel();
		
		tphGraph = new JGraph(tphGraphModel);
		
		tphGraph.setScale(tphScale);
		
		// if there is at leas one property it means that TPH graph exists
		if (model.getPropertiesList().size() > 0)
		{		
			tphCells = null;
		
			// if there is only root node on TPH graph
			if (model.getPropertiesList().size() == 1)
			{
				tphCells = new DefaultGraphCell[3];
			
				Property TPHProperty = model.getPropertiesList().get(0);
		
				tphGraphNodes = new ArrayList<GraphNode>();
			
				GraphNode TPHGraphNode =  new GraphNode(TPHProperty);
			
				GraphConstants.setBounds(TPHGraphNode.getAttributes(), new Rectangle2D.Double());
			
				tphGraphNodes.add(TPHGraphNode);
			
				tphCells[0] = TPHGraphNode;
				GraphConstants.setGradientColor(tphCells[0].getAttributes(), Color.black);
				GraphConstants.setInset(tphCells[0].getAttributes(), 10);
				GraphConstants.setBorderColor(tphCells[0].getAttributes(), Color.black);
				GraphConstants.setAutoSize(tphCells[0].getAttributes(), true);
			
				// create invisible second node
				tphCells[1] = new DefaultGraphCell();
				tphCells[1].addPort();
			
				// create invisible edge between root node and invisible node
				DefaultEdge edge = new DefaultEdge();
				edge.setSource(tphCells[0].getChildAt(0));
				edge.setTarget(tphCells[1].getChildAt(0));
				GraphConstants.setLineColor(edge.getAttributes(), arranger.getTPHBackgroundColor());
				
				tphCells[2] = edge;
			}
			// if there is more than one node on TPH graph
			else
			{
				List<DefaultGraphCell> cellsList = new ArrayList<DefaultGraphCell>();
			
				List<Property> propertiesList = model.getPropertiesList();
			
				tphGraphNodes = new ArrayList<GraphNode>();
			
				// create graph node for each property
				for (Property property : propertiesList)
				{
					GraphNode TPHgraphNode =  new GraphNode(property);
					cellsList.add(TPHgraphNode);
					tphGraphNodes.add(TPHgraphNode);
				}
			
			
				for (GraphNode graphNode : tphGraphNodes)
				{
					List<Property> tranfsormDestinies = model.getTransformDestinies(graphNode.getProperty());
				
					for (GraphNode dependentGraphNode : tphGraphNodes)
					{
						// if some property is destiny property for other property, 
						// then there is edge on graph between that properties 
						if (tranfsormDestinies.contains(dependentGraphNode.getProperty()))
						{
							DefaultEdge edge = new DefaultEdge();
							
							edge.setSource(cellsList.get(tphGraphNodes.indexOf(graphNode)).getChildAt(0));
							edge.setTarget(cellsList.get(tphGraphNodes.indexOf(dependentGraphNode)).getChildAt(0));
						
							if (arranger.isShowFinalizationsAndSplits())
							{
								Object[] label = new Object[1];
						
								if (tranfsormDestinies.size() == 1)
								{
									label[0] = "F";
								}
								else
								{
									label[0] = "S";
								}
						
								Point2D[] labelPositions = {new Point2D.Double(
										GraphConstants.PERMILLE*1/2, -10)};
										GraphConstants.setExtraLabelPositions(edge.getAttributes(),
										labelPositions);
									
								GraphConstants.setExtraLabels(edge.getAttributes(), label);
							}
						
							cellsList.add(edge);
						}
					}
				}
				
			
				// change cells and edges list to table
				tphCells = new DefaultGraphCell[cellsList.size()];
			
				for (int i = 0 ; i < cellsList.size() ; i++)
				{
					tphCells[i] = cellsList.get(i);
				}
			
				// temporary set color of nodes and edges to black 
				for (int i = 0 ; i < tphGraphNodes.size() ; i++)
				{
					GraphConstants.setGradientColor(tphCells[i].getAttributes(), Color.black);
					GraphConstants.setInset(tphCells[i].getAttributes(), 10);
					GraphConstants.setBorderColor(tphCells[i].getAttributes(), Color.black);
					GraphConstants.setAutoSize(tphCells[i].getAttributes(), true);
					GraphConstants.setBounds(tphCells[i].getAttributes(), new Rectangle2D.Double());
				}
			
				for (int i = tphGraphNodes.size() ; i < tphCells.length ; i++)
				{
					GraphConstants.setLineColor(tphCells[i].getAttributes(), Color.BLACK);
					GraphConstants.setLineEnd(tphCells[i].getAttributes(), GraphConstants.ARROW_CLASSIC);
					GraphConstants.setEndFill(tphCells[i].getAttributes(), true);
				}
			}
		
			tphGraph.getGraphLayoutCache().insert(tphCells);
		
			// create invisible root node
			Object roots[] = new Object[1];
			
			roots[0] = tphCells[0];
		
			
			// place nodes in correct layout  
			JGraphFacade facade = new JGraphFacade(tphGraph, roots);
		
			JGraphLayout layout = null;
		
			if (tphLayout == LayoutType.TREE_LAYOUT)
			{
				JGraphTreeLayout graphTreeLayout = new JGraphTreeLayout();
				
				graphTreeLayout.setOrientation(tphOrientation);
				
				if (tphOrientation == SwingConstants.SOUTH) graphTreeLayout.setAlignment(SwingConstants.BOTTOM);
				else graphTreeLayout.setAlignment(SwingConstants.TOP);
				
				graphTreeLayout.setLevelDistance(arranger.getTPHVerticalNodesDistance());
				graphTreeLayout.setNodeDistance(arranger.getTPHHorizontalNodesDistance());
				layout = graphTreeLayout;
			}
		
			if (tphLayout == LayoutType.ORGANIC_LAYOUT)
			{
				com.jgraph.layout.organic.JGraphFastOrganicLayout graphOrganicLayout = new com.jgraph.layout.organic.JGraphFastOrganicLayout();
				graphOrganicLayout.setForceConstant(arranger.getTPHVerticalNodesDistance());
			
				layout = graphOrganicLayout;
			}	
		
			layout.run(facade);
		
			Map nested = facade.createNestedMap(true, true);
		
			tphGraph.getGraphLayoutCache().edit(nested);
			
			for (int i = 0 ; i < tphCells.length ; i++)
			{
				GraphConstants.setSelectable(tphCells[i].getAttributes(), false);
			}
			
			// move graph 50 pixels right and 50 pixelt down
			cds.jlow.client.graph.GraphJ.move(tphGraph, tphCells, 50, 50);
		}
	
		
		tphGraph.setGridEnabled(arranger.isShowGrid());
		tphGraph.setGridVisible(arranger.isShowGrid());
		
		TPHMouseListener tphMouseListener = new TPHMouseListener(model, ARDViewImpl.this);
		
		tphGraph.addMouseListener(tphMouseListener);
		tphGraph.addMouseMotionListener(tphMouseListener);
		tphGraph.addMouseWheelListener(tphMouseListener);
		
		mainFrame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}
	
	
	public void repaintTPHGraph()
	{
		mainFrame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		
		tphPanel.remove(tphScrollPane);
		
		tphGraph.setBackground(arranger.getTPHBackgroundColor());
		
		Map nested=new Hashtable();
		
		
		// map that specifies colors of active nodes
		Map activeMap = new Hashtable();

		if (!arranger.isGradientTPHCell())
		{
			GraphConstants.setGradientColor(activeMap, arranger.getTPHMarkedBackgroundColor());
			GraphConstants.setBackground(activeMap, arranger.getTPHMarkedBackgroundColor());
			GraphConstants.setInset(activeMap, 10);
			GraphConstants.setBorderColor(activeMap, arranger.getTPHMarkedLineColor());
			GraphConstants.setAutoSize(activeMap, true);
			GraphConstants.setOpaque(activeMap, true);
			GraphConstants.setSelectable(activeMap, true);
			GraphConstants.setEditable(activeMap, false);
		}
		else
		{
			GraphConstants.setBackground(activeMap, Color.WHITE);
			GraphConstants.setGradientColor(activeMap, arranger.getTPHMarkedBackgroundColor());
			GraphConstants.setOpaque(activeMap, true);
			GraphConstants.setInset(activeMap, 10);
			GraphConstants.setBorderColor(activeMap, arranger.getTPHMarkedLineColor());
			GraphConstants.setAutoSize(activeMap, true);
			GraphConstants.setSelectable(activeMap, true);
			GraphConstants.setEditable(activeMap, false);
		}
		
		
		// map that specifies colors of inactive nodes
		Map inactiveMap = new Hashtable();

		if (!arranger.isGradientTPHCell())
		{
			GraphConstants.setGradientColor(inactiveMap, arranger.getTPHUnmarkedBackgroundColor());
			GraphConstants.setBackground(inactiveMap, arranger.getTPHUnmarkedBackgroundColor());
			GraphConstants.setInset(inactiveMap, 10);
			GraphConstants.setBorderColor(inactiveMap, arranger.getTPHUnmarkedLineColor());
			GraphConstants.setAutoSize(inactiveMap, true);
			GraphConstants.setOpaque(inactiveMap, true);
			GraphConstants.setSelectable(inactiveMap, true);
			GraphConstants.setEditable(inactiveMap, false);
		}
		else
		{
			GraphConstants.setBackground(inactiveMap, Color.WHITE);
			GraphConstants.setGradientColor(inactiveMap, arranger.getTPHUnmarkedBackgroundColor());
			GraphConstants.setOpaque(inactiveMap, true);
			GraphConstants.setInset(inactiveMap, 10);
			GraphConstants.setBorderColor(inactiveMap, arranger.getTPHUnmarkedLineColor());
			GraphConstants.setAutoSize(inactiveMap, true);
			GraphConstants.setSelectable(inactiveMap, true);
			GraphConstants.setEditable(inactiveMap, false);
		}

		
		// change colors of nodes
		for (int i = 0 ; i < tphCells.length ; i++)
		{
			if (tphCells[i] instanceof GraphNode)
			{
				GraphNode graphNode = (GraphNode)tphCells[i];
				
				if (graphNode.isSelected())
				{
					nested.put(graphNode, activeMap);
				}
				else
				{
					nested.put(graphNode, inactiveMap);
				}
			}
		}
		
		
		// map that specifies colors of active edges
		activeMap = new Hashtable();

		GraphConstants.setLineColor(activeMap, arranger.getTPHMarkedLineColor());
		
		
		// map that specifies colors of inactive edges
		inactiveMap = new Hashtable();

    	GraphConstants.setLineColor(inactiveMap, arranger.getTPHUnmarkedLineColor());

    	
    	// change colors of edges
		for (int i = 0 ; i < tphCells.length ; i++)
		{
			if (tphCells[i] instanceof DefaultEdge)
			{
				DefaultEdge edge = (DefaultEdge)tphCells[i];
				
				DefaultGraphCell sourceCell = (DefaultGraphCell)(((DefaultPort)edge.getSource()).getParent());
				DefaultGraphCell targetCell = (DefaultGraphCell)((DefaultPort)edge.getTarget()).getParent();
				
				if (sourceCell instanceof GraphNode && targetCell instanceof GraphNode)
				{
					if (((GraphNode)sourceCell).isSelected() && ((GraphNode)targetCell).isSelected())
					{
						nested.put(edge, activeMap);
					}
					else
					{
						nested.put(edge, inactiveMap);
					}
				}
			}
		}
		
		
		tphGraph.setGridEnabled(arranger.isShowGrid());
		tphGraph.setGridVisible(arranger.isShowGrid());
		
		
		tphGraph.getGraphLayoutCache().edit(nested,null,null,null);
		
		tphGraph.repaint();
		
		tphScrollPane = new JScrollPane(tphGraph);
		
		tphPanel.add(tphScrollPane, BorderLayout.CENTER);
		
		tphPanel.validate();
		
		if (arranger.isOptimizeGraphsSizes())
		{
			optimizeTPHSize();
		}
		
		mainFrame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}
	
	
	public void refreshTPHGraph()
	{
		createTPHGraph();
		repaintTPHGraph();
	}
	
	
	public void rotateTPHGraph(boolean rotateRight)
	{
		if (rotateRight)
		{
			if (tphOrientation == SwingConstants.WEST) tphOrientation = SwingConstants.NORTH;
			else if (tphOrientation == SwingConstants.NORTH) tphOrientation = SwingConstants.EAST;
			else if (tphOrientation == SwingConstants.EAST) tphOrientation = SwingConstants.SOUTH;
			else if (tphOrientation == SwingConstants.SOUTH) tphOrientation = SwingConstants.WEST;
		}
		else
		{
			if (tphOrientation == SwingConstants.NORTH) tphOrientation = SwingConstants.WEST;
			else if (tphOrientation == SwingConstants.EAST) tphOrientation = SwingConstants.NORTH;
			else if (tphOrientation == SwingConstants.SOUTH) tphOrientation = SwingConstants.EAST;
			else if (tphOrientation == SwingConstants.WEST) tphOrientation = SwingConstants.SOUTH;
		}	
	}
	
	
	public void setTPHLayout(LayoutType layoutType)
	{
		this.tphLayout = layoutType;
	}
	
	
	public void refreshGraphs()
	{
		refreshARDGraph();
		refreshTPHGraph();
	}
	
	
	public void zoomARDGraph(boolean zoomIn)
	{
		if (zoomIn)
		{
			ardScale += 0.1;
		}
		else
		{
			ardScale -= 0.1;
		}
		
		if (ardScale > 1.0)
		{
			ardScale = 1.0;
		}
		if (ardScale < 0.1)
		{
			ardScale = 0.1;
		}
		
		ardGraph.setScale(ardScale);
	}
	

	public void zoomTPHGraph(boolean zoomIn)
	{
		if (zoomIn)
		{
			tphScale += 0.1;
		}
		else
		{
			tphScale -= 0.1;
		}
		
		if (tphScale > 1.0)
		{
			tphScale = 1.0;
		}
		if (tphScale < 0.1)
		{
			tphScale = 0.1;
		}
		
		tphGraph.setScale(tphScale);	
	}
	
	
	public void changeARDNodesDistance(boolean increaseLevelDistance, boolean horizontal)
	{
		if (horizontal)
		{
			int nodesDistance = arranger.getARDHorizontalNodesDistance();
			
			if (increaseLevelDistance)
			{
				nodesDistance += 20;
			}
			else
			{
				nodesDistance -= 20;
			}
			
			if (nodesDistance < 40)
			{
				nodesDistance = 40;
			}
			
			arranger.setARDHorizontalNodesDistance(nodesDistance);
		}
		else
		{
			int nodesDistance = arranger.getARDVerticalNodesDistance();
			
			if (increaseLevelDistance)
			{
				nodesDistance += 20;
			}
			else
			{
				nodesDistance -= 20;
			}
			
			if (nodesDistance < 40)
			{
				nodesDistance = 40;
			}
			
			arranger.setARDVerticalNodesDistance(nodesDistance);
		}
	}
	
	
	public void changeTPHNodesDistance(boolean increaseLevelDistance, boolean horizontal)
	{
		if (horizontal)
		{
			int nodesDistance = arranger.getTPHHorizontalNodesDistance();
			
			if (increaseLevelDistance)
			{
				nodesDistance += 20;
			}
			else
			{
				nodesDistance -= 20;
			}
			
			if (nodesDistance < 40)
			{
				nodesDistance = 40;
			}
			
			arranger.setTPHHorizontalNodesDistance(nodesDistance);
		}
		else
		{
			int nodesDistance = arranger.getTPHVerticalNodesDistance();
			
			if (increaseLevelDistance)
			{
				nodesDistance += 20;
			}
			else
			{
				nodesDistance -= 20;
			}
			
			if (nodesDistance < 40)
			{
				nodesDistance = 40;
			}
			
			arranger.setTPHVerticalNodesDistance(nodesDistance);
		}
	}
	
	
	public JPanel getARDPanel() 
	{
		return ardPanel;
	}


	public JPanel getTPHPanel() 
	{
		return tphPanel;
	}
	
	
	public ArrayList<GraphNode> getARDGraphNodes() 
	{
		return ardGraphNodes;
	}
	
	public void setARDGraphNodes(ArrayList<GraphNode> graphNodes) 
	{
		ardGraphNodes = graphNodes;
	}
	
	
	public ArrayList<GraphNode> getTPHGraphNodes() 
	{
		return tphGraphNodes;
	}
	
	
	public PersistenceManager getPersistenceManager()
	{
		return persistenceManager;
	}
	
	
	public void setARDScale(double scale) 
	{
		ardScale = scale;
	}


	public void setTPHScale(double scale) 
	{
		tphScale = scale;
	}
	
	
	public void dragADRGraph(int x, int y)
	{
		ardScrollPane.getHorizontalScrollBar().setValue(ardScrollPane.getHorizontalScrollBar().getValue() - x);
		ardScrollPane.getVerticalScrollBar().setValue(ardScrollPane.getVerticalScrollBar().getValue() - y);
	}
	
	
	public void dragTPHGraph(int x, int y)
	{
		tphScrollPane.getHorizontalScrollBar().setValue(tphScrollPane.getHorizontalScrollBar().getValue() - x);
		tphScrollPane.getVerticalScrollBar().setValue(tphScrollPane.getVerticalScrollBar().getValue() - y);
	}
	
	
	public void markAllDown(Property property)
	{	
		// mark off all nodes
		for (GraphNode graphNode : ardGraphNodes)
		{
			graphNode.setSelected(false);
		}
		
		for (GraphNode graphNode : tphGraphNodes)
		{
			graphNode.setSelected(false);
		}
		
		
		for (GraphNode graphNode : ardGraphNodes)
		{
			// mark node related to given property
			if (graphNode.getProperty() == property)
			{
				graphNode.setSelected(true);
				
				// mark all nodes that are child nodes to given node 
				recurentMarkDownTPH(graphNode.getProperty());
				
				break;
			}
		}
		
		for (GraphNode graphNode : tphGraphNodes)
		{
			if (graphNode.getProperty() == property)
			{
				graphNode.setSelected(true);
				
				recurentMarkDownTPH(graphNode.getProperty());
				
				break;
			}
		}
		
		repaintARDGraph();
		repaintTPHGraph();
	}
	
	
	/**
	 * Marks all nodes that are related to child properties to given property. Used locally.
	 * 
	 * @param property	property to mark down
	 */
	private void recurentMarkDownTPH(Property property)
	{
		for (GraphNode graphNode : tphGraphNodes)
		{
			if (graphNode.getProperty() == property)
			{
				// mark down given property
				graphNode.setSelected(true);
				
				List<Property> destinyProperties = model.getTransformDestinies(property);
				
				// mark down all child properties to given property
				for (Property destinyProperty : destinyProperties)
				{
					recurentMarkDownTPH(destinyProperty);
				}
			}
		}
		
		for (GraphNode graphNode : ardGraphNodes)
		{
			if (graphNode.getProperty() == property)
			{
				graphNode.setSelected(true);
			}
		}
	}
	
	
	public void rollItem(GraphNode graphNode)
	{
		List<Property> transformSourcesList = model.getTransformSources(graphNode.getProperty());
		
		if (transformSourcesList.size() > 0)
		{
			// get properties that are on the same level than given property (they must be visible after operation)
			Property sourceProperty = transformSourcesList.get(0);
		
			List<Property> transformDestiniesList = model.getTransformDestinies(sourceProperty);
			
			
			// create list of nodes that lays under found properties
			List<GraphNode> removedGraphNodes = new ArrayList<GraphNode>();
		
			for (Property destinyProperty : transformDestiniesList)
			{
				recurentHideDownARD(removedGraphNodes, destinyProperty);
			}
			
			// remove nodes from ARD graph
			ardGraphNodes.removeAll(removedGraphNodes);
			
			
			// if after operation ARD graph doesn't contain node related to given property, create node
			boolean ARDContainSourceProperty = false;
			
			for (GraphNode ARDgraphNode : ardGraphNodes)
			{
				if (ARDgraphNode.getProperty() == sourceProperty)
				{
					ARDContainSourceProperty = true;
				}
			}
			
			if (!ARDContainSourceProperty)
			{
				GraphNode newGraphNode = new GraphNode(sourceProperty);
				GraphConstants.setBounds(newGraphNode.getAttributes(), new Rectangle2D.Double());
				ardGraphNodes.add(newGraphNode);
			}
			
			refreshARDGraph();
			markAllDown(null);
		}
	}
	
	
	/**
	 * Creates list of properties that should be hidden. Used locally.
	 * 
	 * @param removedGraphNodes		list of properties to hide 
	 * 
	 * @param property				actually considered property
	 */
	private void recurentHideDownARD(List<GraphNode> removedGraphNodes, Property property)
	{
		for (GraphNode graphNode : ardGraphNodes)
		{
			if (graphNode.getProperty() == property)
			{
				// add given node to list
				removedGraphNodes.add(graphNode);
				
				break;
			}	
		}
		
		// check all nodes tkat lays under considered node on TPH graph
		List<Property> transformDestinies = model.getTransformDestinies(property);
		
		for (Property destinyProperty : transformDestinies)
		{
			recurentHideDownARD(removedGraphNodes, destinyProperty);
		}
	}
	
	
	public void unrollItem(GraphNode graphNode)
	{
		// get list of properties that are destiny properties to given property
		List<Property> transformDestiniesList = model.getTransformDestinies(graphNode.getProperty());
		
		// remove given property from graph
		ardGraphNodes.remove(graphNode);
		
		// add destiny properties to graph
		for (Property property : transformDestiniesList)
		{
			GraphNode newGraphNode = new GraphNode(property);
			GraphConstants.setBounds(newGraphNode.getAttributes(), new Rectangle2D.Double());
			ardGraphNodes.add(newGraphNode);
		}
		
		refreshARDGraph();
		markAllDown(null);
	}
	
	
	public void unrollAll()
	{
		ardGraphNodes = new ArrayList<GraphNode>();
		
		List<Property> propertiesList = model.getPropertiesList();
		
		for (Property property : propertiesList)
		{
			// add to ARD graph all properties that are leaves of TPH graph
			if (model.getTransformDestinies(property).size() == 0)
			{
				GraphNode newGraphNode = new GraphNode(property);
				GraphConstants.setBounds(newGraphNode.getAttributes(), new Rectangle2D.Double());
				ardGraphNodes.add(newGraphNode);
			}
		}
		
		refreshARDGraph();
		markAllDown(null);
	}
	
	
	public void clearGraphs()
	{
		List<GraphNode> removedTPHGraphNodes = new ArrayList<GraphNode>();
		
		for (GraphNode graphNode : tphGraphNodes)
		{
			if (!model.getPropertiesList().contains(graphNode.getProperty()));
			{
				removedTPHGraphNodes.add(graphNode);
			}
		}
		
		tphGraphNodes.removeAll(removedTPHGraphNodes);
		
		
		ardGraphNodes.removeAll(ardGraphNodes);
		
		List<Property> propertiesList = model.getPropertiesList();
		
		for (Property property : propertiesList)
		{
			if (model.getTransformDestinies(property).size() == 0)
			{
				ardGraphNodes.add(new GraphNode(property));
			}
		}
	}
	
	
	public Arranger getArranger() 
	{
		return arranger;
	}

	
	public JGraph getARDGraph() 
	{
		return ardGraph;
	}


	public void setARDGraph(JGraph graph) 
	{
		ardGraph = graph;
	}
	
	
	public JGraph getTPHGraph() 
	{
		return tphGraph;
	}


	public void setTPHGraph(JGraph graph) 
	{
		tphGraph = graph;
	}
	
	
	public void exportARDToImage()
	{
		JFileChooser fileChooser = new JFileChooser();
		
		fileChooser.setDialogTitle("Export to image");
		
		int returnValue = fileChooser.showDialog(null, "Export");
		
		if (returnValue == JFileChooser.APPROVE_OPTION)
		{
			String filePath = fileChooser.getSelectedFile().getAbsolutePath();
			
			if (!filePath.endsWith(".png"))
			{
				filePath = filePath + ".png";
			}
			
			fileChooser.setVisible(false);
			fileChooser.invalidate();
			
			File file = new File(filePath);
			
			try
			{
				OutputStream out = new FileOutputStream(file);
				
				Color backgroundColor = arranger.getARDBackgroundColor();
				
				double tempScale = ardGraph.getScale();
				
				// high quality image
				ardGraph.setScale(2.0);
				 
				BufferedImage image = ardGraph.getImage(backgroundColor, 50);
				
				ardGraph.setScale(tempScale);
				
				if (image == null)
				{
					image = new BufferedImage(10, 10, BufferedImage.TYPE_INT_RGB);
					Graphics2D g2 = (Graphics2D)image.getGraphics();
					g2.setColor(arranger.getARDBackgroundColor());
					g2.fill(new Rectangle2D.Double(0, 0, image.getWidth(), image.getHeight()));
				}
				
				ImageIO.write(image, "png", out);
				
				out.flush();
				out.close();
			}
			catch (FileNotFoundException e)
			{
				e.printStackTrace();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}
	}
	
	
	public void exportTPHToImage()
	{
		JFileChooser fileChooser = new JFileChooser();
		
		fileChooser.setDialogTitle("Export to image");
		
		int returnValue = fileChooser.showDialog(null, "Export");
		
		if (returnValue == JFileChooser.APPROVE_OPTION)
		{
			String filePath = fileChooser.getSelectedFile().getAbsolutePath();
			
			if (!filePath.endsWith(".png"))
			{
				filePath = filePath + ".png";
			}
			
			fileChooser.setVisible(false);
			fileChooser.invalidate();
			
			File file = new File(filePath);
			
			try
			{
				OutputStream out = new FileOutputStream(file);
				
				Color backgroundColor = arranger.getTPHBackgroundColor();
				
				double tempScale = ardGraph.getScale();
				
				// high quality image
				tphGraph.setScale(2.0);
				
				BufferedImage image = tphGraph.getImage(backgroundColor, 50);
				
				tphGraph.setScale(tempScale);
				
				if (image == null)
				{
					image = new BufferedImage(10, 10, BufferedImage.TYPE_INT_RGB);
					Graphics2D g2 = (Graphics2D)image.getGraphics();
					g2.setColor(arranger.getARDBackgroundColor());
					g2.fill(new Rectangle2D.Double(0, 0, image.getWidth(), image.getHeight()));
				}
				
				ImageIO.write(image, "png", out);
				
				out.flush();
				out.close();
			}
			catch (FileNotFoundException e)
			{
				e.printStackTrace();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}	
		}
	}
	
	
	public void exportBothToImage()
	{
		JFileChooser fileChooser = new JFileChooser();
		
		fileChooser.setDialogTitle("Export to image");
		
		int returnValue = fileChooser.showDialog(null, "Export");
		
		if (returnValue == JFileChooser.APPROVE_OPTION)
		{
			String filePath = fileChooser.getSelectedFile().getAbsolutePath();
			
			String ARDPath = "";
			String TPHPath = "";
			
			if (filePath.endsWith(".png"))
			{
				ARDPath = filePath.substring(0, filePath.lastIndexOf(".png")) + "_ARD.png";
				TPHPath = filePath.substring(0, filePath.lastIndexOf(".png")) + "_TPH.png";
			}
			else
			{
				ARDPath = filePath + "_ARD.png";
				TPHPath = filePath + "_TPH.png";
			}
			
			File ARDfile = new File(ARDPath);
			
			File TPHfile = new File(TPHPath);
			
			try
			{
				OutputStream out = new FileOutputStream(ARDfile);
				
				Color backgroundColor = arranger.getARDBackgroundColor();
				
				double tempScale = ardGraph.getScale();
				
				ardGraph.setScale(2.0);
				
				BufferedImage image = ardGraph.getImage(backgroundColor, 50);
				
				ardGraph.setScale(tempScale);
				
				if (image == null)
				{
					image = new BufferedImage(10, 10, BufferedImage.TYPE_INT_RGB);
					Graphics2D g2 = (Graphics2D)image.getGraphics();
					g2.setColor(arranger.getARDBackgroundColor());
					g2.fill(new Rectangle2D.Double(0, 0, image.getWidth(), image.getHeight()));
				}
				
				ImageIO.write(image, "png", out);
				
				out.flush();
				out.close();
				
				out = new FileOutputStream(TPHfile);
				
				backgroundColor = arranger.getTPHBackgroundColor();
				
				tempScale = tphGraph.getScale();
				
				tphGraph.setScale(2.0);
				
				image = tphGraph.getImage(backgroundColor, 50);
				
				tphGraph.setScale(tempScale);
				
				if (image == null)
				{
					image = new BufferedImage(10, 10, BufferedImage.TYPE_INT_RGB);
					Graphics2D g2 = (Graphics2D)image.getGraphics();
					g2.setColor(arranger.getARDBackgroundColor());
					g2.fill(new Rectangle2D.Double(0, 0, image.getWidth(), image.getHeight()));
				}
				
				ImageIO.write(image, "png", out);
				
				out.flush();
				out.close();
			}
			catch (FileNotFoundException e)
			{
				e.printStackTrace();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}
	}
	

	public void exportToProlog()
	{
		JFileChooser fileChooser = new JFileChooser();
		
		fileChooser.setDialogTitle("Export to image");
		
		int returnValue = fileChooser.showDialog(null, "Export");
		
		if (returnValue == JFileChooser.APPROVE_OPTION)
		{
			String filePath = fileChooser.getSelectedFile().getAbsolutePath();
			
			fileChooser.setVisible(false);
			fileChooser.invalidate();
			
			File file = new File(filePath);
			
			try
			{
				OutputStream fout= new FileOutputStream(file);
				
				OutputStream bout= new BufferedOutputStream(fout);
				
				OutputStreamWriter out = new OutputStreamWriter(bout, "8859_1");
				
				out.write(":- dynamic ard_att/1.\r\n");
				out.write("\r\n");
				
				for (Attribute attribute : model.getAttributesList())
				{
					String attributeContent = "";
					
					if (attribute.getName().trim().toUpperCase().charAt(0) == attribute.getName().trim().charAt(0))
					{
						attributeContent += "'";
					}
					
					attributeContent += attribute.getName().trim();
					
					if (attribute.getName().trim().toUpperCase().charAt(0) == attribute.getName().trim().charAt(0))
					{
						attributeContent += "'";
					}
					
					out.write("ard_att(" + attributeContent + ").\r\n");
				}
				
				out.write("\r\n");
				out.write(":- dynamic ard_property/1.\r\n");
				out.write("\r\n");
				
				
				List<String> propertiesContentList = new ArrayList<String>();
				
				for (Property property : model.getPropertiesList())
				{
					String propertyContent = "[";
					
					for (Attribute propertyAttribute : property.getAttributes())
					{
						String attributeName = propertyAttribute.getName().trim();
						
						if (attributeName.toUpperCase().charAt(0) == attributeName.charAt(0))
						{
							propertyContent += "'";
						}
						
						propertyContent += attributeName;
						
						if (attributeName.toUpperCase().charAt(0) == attributeName.charAt(0))
						{
							propertyContent += "'";
						}
						
						if (propertyAttribute != property.getAttributes().get(property.getAttributes().size() - 1))
						{
							propertyContent += ", ";
						}
					}
					
					propertyContent += "]";
					
					propertiesContentList.add(propertyContent);
				}
				
				for (String propertyContent : propertiesContentList)
				{
					out.write("ard_property(" + propertyContent + ").\r\n");
				}
				
				out.write("\r\n");
				out.write(":- dynamic ard_depend/2.\r\n");
				out.write("\r\n");
				
				for (Dependency dependency : model.getDependenciesList())
				{
					String dependencyContent = "";
					
					dependencyContent += propertiesContentList.get(model.getPropertiesList().indexOf(dependency.getIndependent()));
					
					dependencyContent += ", ";
					
					dependencyContent += propertiesContentList.get(model.getPropertiesList().indexOf(dependency.getDependent()));
					
					out.write("ard_depend(" + dependencyContent + ").\r\n");
				}
				
				out.write("\r\n");
				out.write(":- dynamic ard_hist/2.\r\n");
				out.write("\r\n");
				
				for (Transform transform : model.getTransformsList())
				{
					String transformContent = "";
					
					transformContent += propertiesContentList.get(model.getPropertiesList().indexOf(transform.getSource()));
					
					transformContent += ", ";
					
					transformContent += propertiesContentList.get(model.getPropertiesList().indexOf(transform.getDestiny()));
					
					out.write("ard_hist(" + transformContent + ").\r\n");
				}
				
				out.flush();  
		        out.close();
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
	}
	
	
	public void selectProperty(Property property)
	{
		if (property == null)
		{
			selectedProperty = null;
			
			repaintARDGraph();
			
			return;
		}
		
		// selected property must be simple property with physical attribute
		if (property.getAttributes().size() != 1)
		{
			return;
		}
		
		if (property.getAttributesNames().get(0).toLowerCase().charAt(0) != property.getAttributesNames().get(0).charAt(0))
		{
			return;
		}
		
		// if there is no selected property, select actual property
		if (selectedProperty == null)
		{
			selectedProperty = property;
		}
		// if some property is selected, try to create dependency between selected property and actual property 
		else
		{
			if (addExtraDependency(selectedProperty, property))
			{	
				selectedProperty = null;
				
				refreshGraphs();
			}
		}
		
		repaintARDGraph();
	}
	
	
	public void optimizeARDSize()
	{
		double panelHeight = ardScrollPane.getHeight();
		double panelWidth = ardScrollPane.getWidth();
		
		double graphHeight = ardGraph.getPreferredSize().getHeight() / ardScale;
		double graphWidth = ardGraph.getPreferredSize().getWidth() / ardScale;
		
		double verticalScale = panelHeight / (graphHeight+ 50);
		double horizontalScale = panelWidth / (graphWidth + 50);
		
		double scale = verticalScale;
		if (horizontalScale < verticalScale) scale = horizontalScale;
		
		ardGraph.setScale(scale);
		
		ardScale = scale;
		
		ardGraph.repaint();
		
		ardPanel.remove(ardScrollPane);
		
		ardScrollPane = new JScrollPane(ardGraph);
		
		ardPanel.add(ardScrollPane, BorderLayout.CENTER);
		
		ardPanel.validate();
		
		ardGraph.setMinimumSize(new Dimension(ardGraph.getWidth() + arranger.getARDHorizontalNodesDistance(), ardGraph.getHeight()));
	}
	
	
	public void optimizeTPHSize()
	{
		double panelHeight = tphScrollPane.getHeight();
		double panelWidth = tphScrollPane.getWidth();
		
		double graphHeight = tphGraph.getPreferredSize().getHeight() / tphScale;
		double graphWidth = tphGraph.getPreferredSize().getWidth() / tphScale;
		
		double verticalScale = panelHeight / (graphHeight+ 50);
		double horizontalScale = panelWidth / (graphWidth + 50);
		
		double scale = verticalScale;
		if (horizontalScale < verticalScale) scale = horizontalScale;
		
		tphGraph.setScale(scale);
		
		tphScale = scale;
		
		tphGraph.repaint();
		
		tphPanel.remove(tphScrollPane);
		
		tphScrollPane = new JScrollPane(tphGraph);
		
		tphPanel.add(tphScrollPane, BorderLayout.CENTER);
		
		tphPanel.validate();
		
		tphGraph.setMinimumSize(new Dimension(tphGraph.getWidth() + arranger.getARDHorizontalNodesDistance(), ardGraph.getHeight()));
	}
	
	
	/**
	 * Creates dependency between two properties in result of refactoring 
	 * 
	 * @param independentProperty		independent property
	 * @param dependentProperty			dependent property
	 * 
	 * @return		true if new dependency have been added, false otherwise
	 */
	private boolean addExtraDependency(Property independentProperty, Property dependentProperty)
	{
		List<Dependency> dependenciesList = model.getDependenciesList();
		
		// dependency is added only if there is no dependency with the same independent and dependent properties
		for (Dependency dependency : dependenciesList)
		{
			if (dependency.getIndependent() == independentProperty &&
					dependency.getDependent() == dependentProperty)
			{
				return false;
			}
		}
		
		dependenciesList.add(new Dependency(independentProperty, dependentProperty));
		
		persistenceManager.setStateSaved(false);
		
		return true;
	}
	
	
	/**
	 * Router that sets spline style of edges
	 * 
	 * @author Szymon Książek
	 */
	private class EdgeRouter extends DefaultEdge.LoopRouting 
	{
		private static final long serialVersionUID = 1L;

		public int getEdgeStyle() 
		{
			return GraphConstants.STYLE_SPLINE;
		}
	}
}