/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import java.awt.Color;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.swing.ImageIcon;

/**
 * Implementation of arranger
 * 
 * @author Szymon Książek
 */
public class ArrangerImpl implements Arranger 
{
	/**
	 * Color of marked line on ARD graph
	 */
	private Color ardMarkedLineColor = null;				
	
	/**
	 * Color of unmarked line on ARD graph
	 */
	private Color ardUnmarkedLineColor = null;				
	
	/**
	 * Background color of marked property on ARD graph
	 */
	private Color ardMarkedBackgroundColor = null;			
	
	/**
	 * Background color of unmarked property on ARD graph
	 */
	private Color ardUnmarkedBackgroundColor = null;		
	
	/**
	 * Background color of ARD graph
	 */
	private Color ardBackgroundColor = null;				
	
	/**
	 * Color of marked line on TPH graph
	 */
	private Color tphMarkedLineColor = null;				
	
	/**
	 * Color of unmarked line on TPH graph
	 */
	private Color tphUnmarkedLineColor = null;				
	
	/**
	 * Background color of marked property on TPH graph
	 */
	private Color tphMarkedBackgroundColor = null;			
	
	/**
	 * Background color of unmarked property on TPH graph
	 */
	private Color tphUnmarkedBackgroundColor = null;		
	
	/**
	 * Background color of TPH graph
	 */
	private Color tphBackgroundColor = null;				
	
	/**
	 * If gradient is used to create property's background on ARD graph
	 */
	private boolean isGradientARDCell = false;				
	
	/**
	 * If gradient is used to create property's background on TPH graph
	 */
	private boolean isGradientTPHCell = false;				
	
	/**
	 * Is TPH graph contains letters that indicates finalizations and splits
	 */
	private boolean isShowFinalizationsAndSplits = false;	
	
	/**
	 * true if grid should be visible on graphs, false otherwise
	 */
	private boolean isShowGrid = false;	
	
	/**
	 * Horizontal distance between nodes on ARD graph
	 */
	private int ardHorizontalNodesDistance = 200;
	
	/**
	 * Vertical distance between nodes on ARD graph
	 */
	private int ardVerticalNodesDistance = 100;
	
	/**
	 * Horizontal distance between nodes on TPH graph
	 */
	private int tphHorizontalNodesDistance = 200;
	
	/**
	 * Vertical distance between nodes on TPH graph
	 */
	private int tphVerticalNodesDistance = 200;
	
	/**
	 * Map that contains both icon and icon's names
	 */
	private Map<String, ImageIcon> iconsMap = null;
	
	/**
	 * true to automatically optimize graphs size, false otherwise
	 */
	private boolean optimizeGraphsSizes = false;
	
	
	/**
	 * Constructor
	 */
	public ArrangerImpl()
	{
		ardMarkedLineColor = Color.RED;
		ardUnmarkedLineColor = Color.BLACK;
		ardMarkedBackgroundColor = Color.WHITE;
		ardUnmarkedBackgroundColor = Color.white;
		ardBackgroundColor = Color.WHITE;
		
		tphMarkedLineColor = Color.RED;
		tphUnmarkedLineColor = Color.BLACK;
		tphMarkedBackgroundColor = Color.WHITE;
		tphUnmarkedBackgroundColor = Color.WHITE;
		tphBackgroundColor = Color.WHITE;
		
		isGradientARDCell = false;
		isGradientTPHCell = false;
		
		isShowFinalizationsAndSplits = false;
		
		isShowGrid = false;
		
		ardHorizontalNodesDistance = 200;
		ardVerticalNodesDistance = 100;
		
		tphHorizontalNodesDistance = 200;
		tphVerticalNodesDistance = 200;
		
		optimizeGraphsSizes = false;
		
		
		iconsMap = new HashMap<String, ImageIcon>();
		
		iconsMap.put("add_root", loadIcon("icons/ARDOperation/add_root.gif"));
		iconsMap.put("finalization", loadIcon("icons/ARDOperation/finalization.gif"));
		iconsMap.put("roll", loadIcon("icons/ARDOperation/roll.gif"));
		iconsMap.put("split", loadIcon("icons/ARDOperation/split.gif"));
		iconsMap.put("unroll", loadIcon("icons/ARDOperation/unroll.gif"));
		iconsMap.put("unroll_all", loadIcon("icons/ARDOperation/unroll_all.gif"));
		
		iconsMap.put("organic_layout", loadIcon("icons/GraphSettings/organic_layout.gif"));
		iconsMap.put("rotate_left", loadIcon("icons/GraphSettings/rotate_left.gif"));
		iconsMap.put("rotate_right", loadIcon("icons/GraphSettings/rotate_right.gif"));
		iconsMap.put("tree_layout", loadIcon("icons/GraphSettings/tree_layout.gif"));
		iconsMap.put("zoom_in", loadIcon("icons/GraphSettings/zoom_in.gif"));
		iconsMap.put("zoom_out", loadIcon("icons/GraphSettings/zoom_out.gif"));
		iconsMap.put("increase_horizontal_distance", loadIcon("icons/GraphSettings/increase_horizontal_distance.gif"));
		iconsMap.put("decrease_horizontal_distance", loadIcon("icons/GraphSettings/decrease_horizontal_distance.gif"));
		iconsMap.put("increase_vertical_distance", loadIcon("icons/GraphSettings/increase_vertical_distance.gif"));
		iconsMap.put("decrease_vertical_distance", loadIcon("icons/GraphSettings/decrease_vertical_distance.gif"));
		iconsMap.put("optimize_graph_size", loadIcon("icons/GraphSettings/optimize_graph_size.gif"));
		
		iconsMap.put("independent_windows", loadIcon("icons/MainFrame/independent_windows.gif"));
		iconsMap.put("internal_frames", loadIcon("icons/MainFrame/internal_frames.gif"));
		iconsMap.put("joint_diagram", loadIcon("icons/MainFrame/joint_diagram.gif"));
		iconsMap.put("new", loadIcon("icons/MainFrame/new.gif"));
		iconsMap.put("open", loadIcon("icons/MainFrame/open.gif"));
		iconsMap.put("save", loadIcon("icons/MainFrame/save.gif"));
		iconsMap.put("save_as", loadIcon("icons/MainFrame/save_as.gif"));
		iconsMap.put("split_panes", loadIcon("icons/MainFrame/split_panes.gif"));
		iconsMap.put("tabbed_panes", loadIcon("icons/MainFrame/tabbed_panes.gif"));
		
		iconsMap.put("menu_new", loadIcon("icons/Menu/new.gif"));
		iconsMap.put("menu_open", loadIcon("icons/Menu/open.gif"));
		iconsMap.put("menu_save", loadIcon("icons/Menu/save.gif"));
		iconsMap.put("menu_save_as", loadIcon("icons/Menu/save_as.gif"));
		
		iconsMap.put("add_attribute", loadIcon("icons/TPHOperation/add_attribute.gif"));
		iconsMap.put("add_property", loadIcon("icons/ARDOperation/add_property.gif"));
	}

	
	/**
	 * Loads image icon for given name.
	 * 
	 * @param name	name of icon
	 * 
	 * @return 		image icon associated with given name
	 */
	private ImageIcon loadIcon(String name)
	{
		ImageIcon imageIcon = null;
		
		try
		{
			String jarFilePath = (new File("")).toURI().toURL().toExternalForm() + "ARDEditor.jar";
			URL imgUrl = new URL("jar:" + jarFilePath + "!/images/" + name);
			imageIcon = new ImageIcon(imgUrl);
		}
		catch(MalformedURLException e)
		{
			e.printStackTrace();
		}
	
		if(imageIcon.getIconHeight() == -1)
		{
			return new ImageIcon("images/" + name);
		}
		
		return imageIcon;
	}
	

	public Color getARDMarkedLineColor() 
	{
		return ardMarkedLineColor;
	}

	public void setARDMarkedLineColor(Color markedLineColor) 
	{
		ardMarkedLineColor = markedLineColor;
	}

	public Color getARDUnmarkedLineColor() 
	{
		return ardUnmarkedLineColor;
	}

	public void setARDUnmarkedLineColor(Color unmarkedLineColor) 
	{
		ardUnmarkedLineColor = unmarkedLineColor;
	}

	public Color getARDMarkedBackgroundColor() 
	{
		return ardMarkedBackgroundColor;
	}

	public void setARDMarkedBackgroundColor(Color markedBackgroundColor) 
	{
		ardMarkedBackgroundColor = markedBackgroundColor;
	}

	public Color getARDUnmarkedBackgroundColor() 
	{
		return ardUnmarkedBackgroundColor;
	}

	public void setARDUnmarkedBackgroundColor(Color unmarkedBackgroundColor) 
	{
		ardUnmarkedBackgroundColor = unmarkedBackgroundColor;
	}

	public Color getARDBackgroundColor() 
	{
		return ardBackgroundColor;
	}

	public void setARDBackgroundColor(Color backgroundColor) 
	{
		ardBackgroundColor = backgroundColor;
	}

	public Color getTPHMarkedLineColor() 
	{
		return tphMarkedLineColor;
	}

	public void setTPHMarkedLineColor(Color markedLineColor) 
	{
		tphMarkedLineColor = markedLineColor;
	}

	public Color getTPHUnmarkedLineColor() 
	{
		return tphUnmarkedLineColor;
	}

	public void setTPHUnmarkedLineColor(Color unmarkedLineColor) 
	{
		tphUnmarkedLineColor = unmarkedLineColor;
	}

	public Color getTPHMarkedBackgroundColor() 
	{
		return tphMarkedBackgroundColor;
	}

	public void setTPHMarkedBackgroundColor(Color markedBackgroundColor) 
	{
		tphMarkedBackgroundColor = markedBackgroundColor;
	}

	public Color getTPHUnmarkedBackgroundColor() 
	{
		return tphUnmarkedBackgroundColor;
	}

	public void setTPHUnmarkedBackgroundColor(Color unmarkedBackgroundColor) 
	{
		tphUnmarkedBackgroundColor = unmarkedBackgroundColor;
	}

	public Color getTPHBackgroundColor() 
	{
		return tphBackgroundColor;
	}

	public void setTPHBackgroundColor(Color backgroundColor) 
	{
		tphBackgroundColor = backgroundColor;
	}

	public boolean isGradientARDCell() 
	{
		return isGradientARDCell;
	}

	public void setGradientARDCell(boolean isGradientARDCell) 
	{
		this.isGradientARDCell = isGradientARDCell;
	}

	public boolean isGradientTPHCell() 
	{
		return isGradientTPHCell;
	}

	public void setGradientTPHCell(boolean isGradientTPHCell) 
	{
		this.isGradientTPHCell = isGradientTPHCell;
	}

	public boolean isShowFinalizationsAndSplits() 
	{
		return isShowFinalizationsAndSplits;
	}

	public void setShowFinalizationsAndSplits(boolean isShowFinalizationsAndSplits) 
	{
		this.isShowFinalizationsAndSplits = isShowFinalizationsAndSplits;
	}
	
	public boolean isShowGrid() 
	{
		return isShowGrid;
	}

	public void setShowGrid(boolean isShowGrid) 
	{
		this.isShowGrid = isShowGrid;
	}

	public ImageIcon getIcon(String name)
	{
		return (ImageIcon)iconsMap.get(name);
	}	
	
	public int getARDHorizontalNodesDistance()
	{
		return ardHorizontalNodesDistance;
	}
	
	public void setARDHorizontalNodesDistance(int nodesDistance)
	{
		this.ardHorizontalNodesDistance = nodesDistance;
	}
	
	public int getARDVerticalNodesDistance()
	{
		return ardVerticalNodesDistance;
	}
	
	public void setARDVerticalNodesDistance(int nodesDistance)
	{
		this.ardVerticalNodesDistance = nodesDistance;
	}
	
	public int getTPHHorizontalNodesDistance()
	{
		return tphHorizontalNodesDistance;
	}
	
	public void setTPHHorizontalNodesDistance(int nodesDistance)
	{
		this.tphHorizontalNodesDistance = nodesDistance;
	}
	
	public int getTPHVerticalNodesDistance()
	{
		return tphVerticalNodesDistance;
	}
	
	public void setTPHVerticalNodesDistance(int nodesDistance)
	{
		this.tphVerticalNodesDistance = nodesDistance;
	}
	
	public void setOptimizeGraphsSizes(boolean optimize)
	{
		this.optimizeGraphsSizes = optimize;
	}
	
	public boolean isOptimizeGraphsSizes()
	{
		return optimizeGraphsSizes;
	}
}