/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

import pl.edu.agh.hekate.ard.actions.*;
import pl.edu.agh.hekate.ard.dialogs.AboutEditorDialog;

/**
 * Main frame of application
 * 
 * @author Szymon Książek
 */
public class MainFrame extends JFrame
{
	private static final long serialVersionUID = 1L;
	
	/**
	 * reference to model instance
	 */
	private ARDModel mainARDModel = null;

	/**
	 * reference to view instance
	 */
	private ARDView mainARDView = null;
	
	/**
	 * reference to persistence manager instance
	 */
	private PersistenceManager persistenceManager = null;
	
	/**
	 * composition type of graphs
	 */
	private CompositionType compositionType = null;			
	
	/**
	 * used with split panes composition type
	 */
	private JSplitPane splitPane = null; 
	
	/**
	 * used with tabbed panes composition type
	 */
	private JTabbedPane tabbedPane = null;
	
	/**
	 * used with internal frames composition type
	 */
	private JDesktopPane desktopPane = null;
	
	
	/**
	 * used to represent ARD graph with independent windows composition type
	 */
	private JFrame ARDWindow = null;
	
	/**
	 * used to represent TPH graph with independent windows composition type
	 */
	private JFrame TPHWindow = null;
	
	/**
	 * menu check box specifies that split panes are in use
	 */
	private JCheckBoxMenuItem menuSplitPanes = null;
	
	/**
	 * menu check box specifies that tabbed panes are in use
	 */
	private JCheckBoxMenuItem menuTabbedPanes = null;
	
	/**
	 * menu check box specifies that internal frames are in use
	 */
	private JCheckBoxMenuItem menuInternalFrames = null;
	
	/**
	 * menu check box specifies that independent windows are in use
	 */
	private JCheckBoxMenuItem menuIndependentWindows = null;
	
	
	/**
	 * Constructor
	 */
	public MainFrame()
	{
		setTitle("HJed");
			
		int screenWidth = (int)getContentPane().getToolkit().getScreenSize().getWidth();
		int screenHeight = (int)getContentPane().getToolkit().getScreenSize().getHeight();
			
		setLocation(100, 50);
			
		setSize(screenWidth - 200, screenHeight - 100);
		
		// create model instance
		mainARDModel = new ARDModelImpl();
		
		// cerate persistence manager instance
		persistenceManager = new PersistenceManagerImpl(mainARDModel, mainARDView, MainFrame.this);
		
		// create view instance
		mainARDView = new ARDViewImpl(mainARDModel, persistenceManager, MainFrame.this);
		
		persistenceManager.setView(mainARDView);
		
		splitPane = null;
		tabbedPane = null;
		desktopPane =  null;
		ARDWindow = null;
		TPHWindow = null;
	
		// set default composition type to split panes
		compositionType = CompositionType.SPLIT_PANES;
		
		changeCompositionType(compositionType);
		
		// create actions to change composition type
		Action setSplitPanesAction = new CompositionTypeChangeAction("Split panes", mainARDView.getArranger().getIcon("split_panes"), CompositionType.SPLIT_PANES, MainFrame.this);
		Action setTabbedPanesAction = new CompositionTypeChangeAction("Tabbed panes", mainARDView.getArranger().getIcon("tabbed_panes"), CompositionType.TABBED_PANES, MainFrame.this);
		Action setInternalFramesAction = new CompositionTypeChangeAction("Internal frames", mainARDView.getArranger().getIcon("internal_frames"), CompositionType.INTERNAL_FRAMES, MainFrame.this);
		Action setIndependentWindowsAction = new CompositionTypeChangeAction("Independent windows", mainARDView.getArranger().getIcon("independent_windows"), CompositionType.INDEPENDENT_WINDOWS, MainFrame.this);
		
		// create action to show joint diagram
		Action showJointDiagram = new JointDiagramAction("Show joint diagram", mainARDView.getArranger().getIcon("joint_diagram"), MainFrame.this);
		
		// create actions to open, save or create new diagrams
		Action newAction = new NewAction("New", mainARDView.getArranger().getIcon("new"), mainARDModel, mainARDView, persistenceManager, MainFrame.this);
		Action openAction = new OpenAction("Open", mainARDView.getArranger().getIcon("open"), mainARDView, persistenceManager);
		Action saveAsAction = new SaveAsAction("Save as", mainARDView.getArranger().getIcon("save_as"), persistenceManager);
		Action saveAction = new SaveAction("Save", mainARDView.getArranger().getIcon("save"), persistenceManager);
		
		
		// create panel for tool bars
		JPanel toolBarsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		
		
		// create tool bar for new, save and open operations
		JToolBar mainOperationsTypeToolBar = new JToolBar();
		
		mainOperationsTypeToolBar.add(newAction);
		mainOperationsTypeToolBar.add(openAction);
		mainOperationsTypeToolBar.add(saveAsAction);
		mainOperationsTypeToolBar.add(saveAction);
		
		mainOperationsTypeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		mainOperationsTypeToolBar.setName("Main operations");
		
		toolBarsPanel.add(mainOperationsTypeToolBar);
		
		
		// create tool bar for operations responsible for change composition type and show joint diagram
		JToolBar compositionTypeToolBar = new JToolBar();
		
		compositionTypeToolBar.add(setSplitPanesAction);
		compositionTypeToolBar.add(setTabbedPanesAction);
		compositionTypeToolBar.add(setInternalFramesAction);
		compositionTypeToolBar.add(setIndependentWindowsAction);
		
		compositionTypeToolBar.addSeparator();
		compositionTypeToolBar.add(showJointDiagram);
		
		compositionTypeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		compositionTypeToolBar.setName("Composition type");
		
		toolBarsPanel.add(compositionTypeToolBar);
		
	
		// add tool bars to window
		add(toolBarsPanel, BorderLayout.NORTH);
		
		
		// create menu
		JMenuBar menuBar = new JMenuBar();
		
		JMenu menuFile = new JMenu("File");
		menuFile.setMnemonic('F');
		
		JMenuItem menuNew = new JMenuItem(new NewAction("New", mainARDView.getArranger().getIcon("menu_new"), mainARDModel, mainARDView, persistenceManager, MainFrame.this));
		menuNew.setMnemonic('N');
		
		JMenuItem menuOpen = new JMenuItem(new OpenAction("Open", mainARDView.getArranger().getIcon("menu_open"), mainARDView, persistenceManager));
		menuOpen.setMnemonic('O');
		
		JMenuItem menuSaveAs = new JMenuItem(new SaveAsAction("Save as", mainARDView.getArranger().getIcon("menu_save_as"), persistenceManager));
		menuSaveAs.setMnemonic('a');
		
		JMenuItem menuSave = new JMenuItem(new SaveAction("Save", mainARDView.getArranger().getIcon("menu_save"), persistenceManager));
		menuSave.setMnemonic('S');
		
		JCheckBoxMenuItem menuAutosave = new JCheckBoxMenuItem("Autosave");
		menuAutosave.setMnemonic('u');
		
		menuAutosave.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton)event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							persistenceManager.setAutosave(true);
							
							if (!persistenceManager.isAutosave())
							{
								aButton.getModel().setSelected(false);
							}
						}
						else
						{
							persistenceManager.setAutosave(false);
						}
					}
				});
		
		
		JMenu menuExportToImage = new JMenu("Export to image");
		menuExportToImage.setMnemonic('E');
		
		JMenuItem menuExportARDToImage = new JMenuItem("Export ARD");
		menuExportARDToImage.setMnemonic('A');
		
		menuExportARDToImage.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						mainARDView.exportARDToImage();
					}
				});
		
		JMenuItem menuExportTPHToImage = new JMenuItem("Export TPH");
		menuExportTPHToImage.setMnemonic('T');
		
		menuExportTPHToImage.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						mainARDView.exportTPHToImage();
					}
				});
		
		
		JMenuItem menuExportBothToImage = new JMenuItem("Export both");
		menuExportBothToImage.setMnemonic('b');
		
		menuExportBothToImage.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						mainARDView.exportBothToImage();
					}
				});
		
		
		menuExportToImage.add(menuExportARDToImage);
		menuExportToImage.add(menuExportTPHToImage);
		menuExportToImage.addSeparator();
		menuExportToImage.add(menuExportBothToImage);
		
		
		JMenuItem menuExportToProlog = new JMenuItem("Export to Prolog");
		menuExportToProlog.setMnemonic('P');
		
		menuExportToProlog.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						mainARDView.exportToProlog();
					}
				});
		
		
		// create action responsible for closing application
		ExitAction exitAction = new ExitAction("Exit", null, persistenceManager, MainFrame.this);
		
		JMenuItem menuExit = new JMenuItem(exitAction);
		menuExit.setMnemonic('x');
		
		menuFile.add(menuNew);
		menuFile.addSeparator();
		menuFile.add(menuOpen);
		menuFile.addSeparator();
		menuFile.add(menuSave);
		menuFile.add(menuSaveAs);
		menuFile.addSeparator();
		menuFile.add(menuAutosave);
		menuFile.addSeparator();
		menuFile.add(menuExportToProlog);
		menuFile.addSeparator();
		menuFile.add(menuExportToImage);
		menuFile.addSeparator();
		menuFile.add(menuExit);
		
		menuBar.add(menuFile);
		
		
		// create "Refactor" menu 
		JMenu menuRefactor = new JMenu("Refactor");
		menuRefactor.setMnemonic('R');
		
		JMenuItem menuRefactorAttributes = new JMenuItem(new RefactorAttributesAction("Attributes", null, mainARDModel, mainARDView));
		menuRefactorAttributes.setMnemonic('A');
		
		menuRefactor.add(menuRefactorAttributes);
		
		menuBar.add(menuRefactor);
		
		
		// create "View" menu
		JMenu menuView = new JMenu("View");
		menuView.setMnemonic('V');
		
		menuSplitPanes = new JCheckBoxMenuItem("Split panes");
		menuSplitPanes.setMnemonic('S');
		
		menuSplitPanes.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							changeCompositionType(CompositionType.SPLIT_PANES);
						}
					}
				});
		
		menuSplitPanes.setSelected(true);
		
		
		menuTabbedPanes = new JCheckBoxMenuItem("Tabbed panes");
		menuTabbedPanes.setMnemonic('T');
		
		menuTabbedPanes.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							changeCompositionType(CompositionType.TABBED_PANES);
						}
					}
				});
		
		
		menuInternalFrames = new JCheckBoxMenuItem("Internal frames");
		menuInternalFrames.setMnemonic('I');
		
		menuInternalFrames.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							changeCompositionType(CompositionType.INTERNAL_FRAMES);
						}
						else
						{
							
						}
					}
				});
		
		
		menuIndependentWindows = new JCheckBoxMenuItem("Independent windows");
		menuIndependentWindows.setMnemonic('n');
		
		menuIndependentWindows.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							changeCompositionType(CompositionType.INDEPENDENT_WINDOWS);
						}
						else
						{
							
						}
					}
				});
		
		
		
		ButtonGroup compositionTypeButonGroup = new ButtonGroup();
		
		compositionTypeButonGroup.add(menuSplitPanes);
		compositionTypeButonGroup.add(menuTabbedPanes);
		compositionTypeButonGroup.add(menuInternalFrames);
		compositionTypeButonGroup.add(menuIndependentWindows);
		
		
		menuView.add(menuSplitPanes);
		menuView.add(menuTabbedPanes);
		menuView.add(menuInternalFrames);
		menuView.add(menuIndependentWindows);
		
		
		JCheckBoxMenuItem menuMarkFinalizationsAndSplits = new JCheckBoxMenuItem("Mark finalizations and splits");
		menuMarkFinalizationsAndSplits.setMnemonic('M');
		
		menuMarkFinalizationsAndSplits.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							mainARDView.getArranger().setShowFinalizationsAndSplits(true);
							mainARDView.refreshTPHGraph();
						}
						else
						{
							mainARDView.getArranger().setShowFinalizationsAndSplits(false);
							mainARDView.refreshTPHGraph();
						}
					}
				});
		
		menuView.addSeparator();
		menuView.add(menuMarkFinalizationsAndSplits);
		
		
		JCheckBoxMenuItem menuShowGrid = new JCheckBoxMenuItem("Show grid");
		menuShowGrid.setMnemonic('g');
		
		menuShowGrid.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							mainARDView.getArranger().setShowGrid(true);
							mainARDView.repaintARDGraph();
							mainARDView.repaintTPHGraph();
						}
						else
						{
							mainARDView.getArranger().setShowGrid(false);
							mainARDView.repaintARDGraph();
							mainARDView.repaintTPHGraph();
						}
					}
				});
		
		menuView.addSeparator();
		menuView.add(menuShowGrid);
		
		
		
		JCheckBoxMenuItem optimizeGraphsSizes = new JCheckBoxMenuItem("Optimize graphs sizes");
		optimizeGraphsSizes.setMnemonic('O');
		
		optimizeGraphsSizes.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AbstractButton aButton = (AbstractButton) event.getSource();
						
						if (aButton.getModel().isSelected())
						{
							mainARDView.getArranger().setOptimizeGraphsSizes(true);
							mainARDView.optimizeARDSize();
							mainARDView.optimizeTPHSize();
						}
						else
						{
							mainARDView.getArranger().setOptimizeGraphsSizes(true);
							mainARDView.getArranger().setShowGrid(false);
						}
					}
				});
		
		menuView.addSeparator();
		menuView.add(optimizeGraphsSizes);
		
		
		
		
		JMenuItem menuJointDiagram = new JMenuItem(new JointDiagramAction("Show joint diagram", null, MainFrame.this));
		menuJointDiagram.setMnemonic('j');
		
		menuView.addSeparator();
		menuView.add(menuJointDiagram);
		
		
		JMenuItem menuSetColors = new JMenuItem(new SetColorsAction("Set colors", null, mainARDView));
		menuSetColors.setMnemonic('c');
		
		menuView.addSeparator();
		menuView.add(menuSetColors);
		
		menuBar.add(menuView);
		
		// create "Help" menu
		JMenu menuHelp = new JMenu("Help");
		menuHelp.setMnemonic('H');
		
		JMenuItem menuAboutEditor = new JMenuItem("About HJed");
		menuAboutEditor.setMnemonic('A');
		
		menuAboutEditor.addActionListener(new
				ActionListener()
				{
					public void actionPerformed(ActionEvent event)
					{
						AboutEditorDialog aboutEditorDialog = new AboutEditorDialog();
						aboutEditorDialog.showDialog("About HJed");
					}
				});
		
		menuHelp.add(menuAboutEditor);
		
		menuBar.add(menuHelp);
		
		
		setJMenuBar(menuBar);
		
		// because ExitAction is responsible for closing window
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		
		MainFrame.this.addWindowListener(new CloseWindowAction(exitAction));
	}
	
	
	/**
	 * Changes composition type
	 * 
	 * @param newCompositionType	new composition type
	 */
	public void changeCompositionType(CompositionType newCompositionType)
	{
		// remove actual composition type
		if (compositionType == CompositionType.SPLIT_PANES)
		{
			if(splitPane != null)
			{
				remove(splitPane);
				splitPane = null;
			}
		}
		
		if (compositionType == CompositionType.TABBED_PANES)
		{
			if(tabbedPane != null)
			{
				remove(tabbedPane);
				tabbedPane = null;
			}
		}
		
		if (compositionType == CompositionType.INTERNAL_FRAMES)
		{
			if(desktopPane != null)
			{
				remove(desktopPane);
				desktopPane = null;
			}
		}
		
		if (compositionType == CompositionType.INDEPENDENT_WINDOWS)
		{
			if(ARDWindow != null)
			{
				ARDWindow.setVisible(false);
				ARDWindow = null;
			}
			
			if(TPHWindow != null)
			{
				TPHWindow.setVisible(false);
				TPHWindow = null;
			}
		}
		
		
		// create new composition type
		if(newCompositionType == CompositionType.SPLIT_PANES)
		{
			splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, mainARDView.getARDPanel(), mainARDView.getTPHPanel());
			
			splitPane.setDividerLocation((int)(MainFrame.this.getWidth() * 0.5));

			add(splitPane, BorderLayout.CENTER);
			
			if (menuSplitPanes != null) menuSplitPanes.setSelected(true);
		}
		
		if(newCompositionType == CompositionType.TABBED_PANES)
		{
			tabbedPane = new JTabbedPane();
			
			tabbedPane.add("ARD area", mainARDView.getARDPanel());
			tabbedPane.add("TPH area", mainARDView.getTPHPanel());

			add(tabbedPane, BorderLayout.CENTER);
			
			if (menuTabbedPanes != null) menuTabbedPanes.setSelected(true);
		}
		
		if(newCompositionType == CompositionType.INTERNAL_FRAMES)
		{
			desktopPane = new JDesktopPane();
			
			desktopPane.setBackground(MainFrame.this.getBackground());
			
			add(desktopPane, BorderLayout.CENTER);
			
			JInternalFrame ARDFrame = new JInternalFrame("ARD area", true, false, true, true);
			JInternalFrame TPHFrame = new JInternalFrame("TPH area", true, false, true, true);
			
			ARDFrame.getContentPane().add(mainARDView.getARDPanel());
			TPHFrame.getContentPane().add(mainARDView.getTPHPanel());
			
			desktopPane.add(ARDFrame);
			desktopPane.add(TPHFrame);
			
			MainFrame.this.validate();
			
			ARDFrame.reshape(0, 0, desktopPane.getWidth() / 2, desktopPane.getHeight());
			TPHFrame.reshape(desktopPane.getWidth() / 2, 0, desktopPane.getWidth() / 2, desktopPane.getHeight());
			
			ARDFrame.setVisible(true);
			TPHFrame.setVisible(true);
			
			if (menuInternalFrames != null) menuInternalFrames.setSelected(true);
		}
		
		if(newCompositionType == CompositionType.INDEPENDENT_WINDOWS)
		{
			ARDWindow = new JFrame("ARD area");
			ARDWindow.add(mainARDView.getARDPanel());
			ARDWindow.setBounds(MainFrame.this.getX() + 20, MainFrame.this.getY() + 100, MainFrame.this.getWidth() / 2 - 30, MainFrame.this.getHeight() - 120);
			ARDWindow.setMinimumSize(new Dimension(300, 300));
			ARDWindow.setVisible(true);
			
			TPHWindow = new JFrame("TPH area");
			TPHWindow.add(mainARDView.getTPHPanel());
			TPHWindow.setBounds(MainFrame.this.getX() + MainFrame.this.getWidth() / 2 + 10, MainFrame.this.getY() + 100, MainFrame.this.getWidth() / 2 - 30, MainFrame.this.getHeight() - 120);
			TPHWindow.setMinimumSize(new Dimension(300, 300));
			TPHWindow.setVisible(true);
			
			if (menuIndependentWindows != null) menuIndependentWindows.setSelected(true);
		}
		
		compositionType = newCompositionType;
		
		mainARDView.refreshGraphs();
		
		MainFrame.this.repaint();
		MainFrame.this.validate();
	}
	
	
	/**
	 * Returns composition type of application
	 * 
	 * @return	composition type of application
	 */
	public CompositionType getCompositionType()
	{
		return compositionType;
	}


	/**
	 * Returns model instance
	 * 
	 * @return	actual instance of model
	 */
	public ARDModel getMainARDModel() 
	{
		return mainARDModel;
	}


	/**
	 * Sets model of application
	 * 
	 * @param mainARDModel	new model of application
	 */
	public void setMainARDModel(ARDModel mainARDModel) 
	{
		this.mainARDModel = mainARDModel;
	}


	/**
	 * Returns view instance
	 * 
	 * @return	actual instance of view
	 */
	public ARDView getMainARDView() 
	{
		return mainARDView;
	}


	/**
	 * Sets view of application
	 * 
	 * @param mainARDView	new view of application
	 */
	public void setMainARDView(ARDView mainARDView) 
	{
		this.mainARDView = mainARDView;
	}
}
