/*
 *     Implementation by Szymon Ksiazek <szymonh20@poczta.fm>
 *
 *     Copyright (C) 2006-9 by the HeKatE Project
 *
 *     HJEd has been develped within the HeKatE Project, 
 *     see http://hekate.ia.agh.edu.pl
 *
 *     This file is part of HJEd.
 *
 *     HJEd is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HJEd is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HJEd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package pl.edu.agh.hekate.ard.dialogs;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.SwingConstants;

import org.jgraph.JGraph;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.GraphConstants;

import pl.edu.agh.hekate.ard.ARDModel;
import pl.edu.agh.hekate.ard.ARDMouseListener;
import pl.edu.agh.hekate.ard.ARDView;
import pl.edu.agh.hekate.ard.ARDViewImpl;
import pl.edu.agh.hekate.ard.Arranger;
import pl.edu.agh.hekate.ard.CompositionType;
import pl.edu.agh.hekate.ard.GraphNode;
import pl.edu.agh.hekate.ard.LayoutType;
import pl.edu.agh.hekate.ard.MainFrame;
import pl.edu.agh.hekate.ard.Property;
import pl.edu.agh.hekate.ard.actions.AddFinalizationAction;
import pl.edu.agh.hekate.ard.actions.AddRootNodeAction;
import pl.edu.agh.hekate.ard.actions.AddSplitAction;
import pl.edu.agh.hekate.ard.actions.ChangeNodesDistanceAction;
import pl.edu.agh.hekate.ard.actions.LayoutTypeChangeAction;
import pl.edu.agh.hekate.ard.actions.OptimizeDiagramSizeAction;
import pl.edu.agh.hekate.ard.actions.RollAction;
import pl.edu.agh.hekate.ard.actions.RotateGraphAction;
import pl.edu.agh.hekate.ard.actions.ZoomGraphAction;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.JGraphLayout;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;
import com.jgraph.layout.organic.JGraphFastOrganicLayout;
import com.jgraph.layout.tree.JGraphTreeLayout;

/**
 * Dialog that shows joint diagram
 * 
 * @author Szymon Książek
 *
 */
public class JointDiagramDialog extends JPanel
{
	private static final long serialVersionUID = 1L;

	/**
	 * reference to main frame of application
	 */
	private MainFrame mainFrame;
	
	/**
	 * view reference
	 */
	private ARDView view = null;
	
	/**
	 * arranger reference
	 */
	private Arranger arranger = null;
	
	/**
	 * JGraph instance
	 */
	private JGraph graph = null;
	
	/**
	 * scroll pane that contains graph
	 */
	private JScrollPane graphScrollPane;
	
	/**
	 * graph's cells
	 */
	private DefaultGraphCell[] cells = null;
	
	/**
	 * true if tree layout is used, false if organic layout
	 */
	private boolean treeLayout = true;
	
	/**
	 * indicates graph's orientation
	 */
	private int graphOrientation = SwingConstants.WEST;
	
	/**
	 * scale of graph
	 */
	private double graphScale = 0.4;
	
	/**
	 * horizontal distance between graph's nodes
	 */
	private int horizontalNodesDistance = 200;
	
	/**
	 * vertical distance between graph's nodes
	 */
	private int verticalNodesDistance = 200;
	
	/**
	 * dialog reference
	 */
	private JFrame dialog = null;	
	
	
	
	/**
	 * Constructor
	 * 
	 * @param mainFrame	reference to main frame of application
	 */
	public JointDiagramDialog(MainFrame mainFrame)
	{
		this.mainFrame = mainFrame;
		this.view = mainFrame.getMainARDView();
		this.arranger = view.getArranger();
		
		setLayout(new BorderLayout());
		
		createGraph();
		
		graphScrollPane = new JScrollPane(graph);
		
		add(graphScrollPane);
		
		JPanel buttonPanel = new JPanel();

		// button that closes dialog
		JButton closeButton = new JButton("Close");
		
		closeButton.addActionListener(new
				ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				dialog.setVisible(false);
			}
		});
		
		buttonPanel.add(closeButton);
		
		add(buttonPanel, BorderLayout.SOUTH);
	}
	
	
	/**
	 * Method used to create joint graph
	 * 
	 * @return joint graph
	 */
	private void createGraph()
	{	
		if (dialog == null)
		{	
			mainFrame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		}
		else
		{	
			dialog.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		}
		
		ARDModel model = mainFrame.getMainARDModel();
		
		DefaultGraphModel graphModel = new DefaultGraphModel();
		
		graph = new JGraph(graphModel);
		
		graph.setGridEnabled(arranger.isShowGrid());
		graph.setGridVisible(arranger.isShowGrid());
		
		graph.setScale(graphScale);
		
		if (model.getPropertiesList().size() > 0)
		{
			cells = null;
		
			// if there is only one property in model
			if (model.getPropertiesList().size() == 1)
			{
				cells = new DefaultGraphCell[3];
			
				Property property = model.getPropertiesList().get(0);
				GraphNode graphNode =  new GraphNode(property);
			
				GraphConstants.setInset(graphNode.getAttributes(), 10);
				GraphConstants.setBackground(graphNode.getAttributes(), new Color(200,200,255));
				GraphConstants.setBorderColor(graphNode.getAttributes(), Color.BLUE);
				GraphConstants.setAutoSize(graphNode.getAttributes(), true);
				GraphConstants.setOpaque(graphNode.getAttributes(), true);
				
				cells[0] = graphNode;
				
				GraphConstants.setSelectable(cells[0].getAttributes(), false);
			
				cells[1] = new DefaultGraphCell();
				cells[1].addPort();
			
				DefaultEdge edge = new DefaultEdge();
				edge.setSource(cells[0].getChildAt(0));
				edge.setTarget(cells[1].getChildAt(0));
				GraphConstants.setLineColor(edge.getAttributes(), Color.WHITE);
				cells[2] = edge;
			}		
			// if there are more than one property in model
			else
			{
				List<DefaultGraphCell> cellsList = new ArrayList<DefaultGraphCell>();
			
				// get all properties
				List<Property> propertiesList = model.getPropertiesList();
			
				// create graph node for each one property
				List<GraphNode> graphNodes = new ArrayList<GraphNode>();
			
				for (Property property : propertiesList)
				{
					GraphNode graphNode =  new GraphNode(property);
					
					GraphConstants.setInset(graphNode.getAttributes(), 10);
					GraphConstants.setBorderColor(graphNode.getAttributes(), Color.BLACK);
					GraphConstants.setAutoSize(graphNode.getAttributes(), true);
					GraphConstants.setSelectable(graphNode.getAttributes(), true);
					GraphConstants.setBounds(graphNode.getAttributes(), new Rectangle2D.Double());
					
					cellsList.add(graphNode);
					graphNodes.add(graphNode);
				}
			
			
				// create edges that represent transforms
				for (GraphNode graphNode : graphNodes)
				{
					List<Property> tranfsormSources = model.getTransformDestinies(graphNode.getProperty());
				
					for (GraphNode dependentGraphNode : graphNodes)
					{
						if (tranfsormSources.contains(dependentGraphNode.getProperty()))
						{
							DefaultEdge edge = new DefaultEdge();
							edge.setSource(cellsList.get(graphNodes.indexOf(graphNode)).getChildAt(0));
							edge.setTarget(cellsList.get(graphNodes.indexOf(dependentGraphNode)).getChildAt(0));
							
							GraphConstants.setLineColor(edge.getAttributes(), Color.BLACK);
							GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_CLASSIC);
							GraphConstants.setEndFill(edge.getAttributes(), true);
							GraphConstants.setSelectable(edge.getAttributes(), false);
							GraphConstants.setRouting(edge.getAttributes(), new EdgeRouter());
							
							//GraphConstants.setRouting(edge.getAttributes(), GraphConstants.ROUTING_DEFAULT);
							
							cellsList.add(edge);
						}
					}
				}
				
				
				// at the moment the TPH graph is created
				
				// create edges that represents dependencies between properties
				for (Property property : propertiesList)
				{
					// if property has not child properties (it is property from the last phase of design), mark it with
					// blue color
					if (model.getTransformDestinies(property).size() == 0)
					{
						GraphConstants.setInset(cellsList.get(propertiesList.indexOf(property)).getAttributes(), 10);
						GraphConstants.setBackground(cellsList.get(propertiesList.indexOf(property)).getAttributes(), new Color(200, 200, 255));
						GraphConstants.setBorderColor(cellsList.get(propertiesList.indexOf(property)).getAttributes(), Color.BLUE);
						GraphConstants.setAutoSize(cellsList.get(propertiesList.indexOf(property)).getAttributes(), true);
						GraphConstants.setOpaque(cellsList.get(propertiesList.indexOf(property)).getAttributes(), true);
						
						// mark all lines that represent connections on most detailed ARD graph
						for (Property dependentProperty : model.getDependentProperties(property))
						{
							if (model.getTransformDestinies(dependentProperty).size() == 0)
							{
								DefaultEdge edge = new DefaultEdge();
								edge.setSource(cellsList.get(propertiesList.indexOf(property)).getChildAt(0));
								edge.setTarget(cellsList.get(propertiesList.indexOf(dependentProperty)).getChildAt(0));
							
								GraphConstants.setLineColor(edge.getAttributes(), Color.BLUE);
								GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_CLASSIC);
								GraphConstants.setEndFill(edge.getAttributes(), true);
								GraphConstants.setSelectable(edge.getAttributes(), false);
							
								GraphConstants.setRouting(edge.getAttributes(), new EdgeRouter());
							
								cellsList.add(edge);
							}
						}
					}
				}
		
				cells = new DefaultGraphCell[cellsList.size()];
			
				for (int i = 0 ; i < cellsList.size() ; i++)
				{
					cells[i] = cellsList.get(i);
				}		
			}
		
			graph.getGraphLayoutCache().insert(cells);
		
		
			JGraphFacade facade = new JGraphFacade(graph);
		
			
			// sets graph's layout
			JGraphLayout layout = null;
			
			if (treeLayout)
			{
			
				JGraphHierarchicalLayout graphHierarchicalLayout = new JGraphHierarchicalLayout();
				
				graphHierarchicalLayout.setOrientation(graphOrientation);
				graphHierarchicalLayout.setFineTuning(true);
				graphHierarchicalLayout.setFixRoots(true);
				graphHierarchicalLayout.setInterRankCellSpacing(horizontalNodesDistance);
				graphHierarchicalLayout.setIntraCellSpacing(verticalNodesDistance);
				
				
				layout = graphHierarchicalLayout;
			}
			else
			{
				JGraphFastOrganicLayout graphOrganicLayout = new JGraphFastOrganicLayout();
				
				graphOrganicLayout.setForceConstant(verticalNodesDistance);
				
				layout = graphOrganicLayout;
			}
			
			layout.run(facade);
		
			Map nested = facade.createNestedMap(true, true);
		
			graph.getGraphLayoutCache().edit(nested);
		
			cds.jlow.client.graph.GraphJ.move(graph, cells, 50, 50);
			
			MouseListener mouseListener = new MouseListener();
			
			graph.addMouseListener(mouseListener);
			graph.addMouseMotionListener(mouseListener);
			graph.addMouseWheelListener(mouseListener);
		}
		
		if (dialog == null)
		{	
			mainFrame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		}
		else
		{	
			dialog.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		}
	}
	
	
	/**
	 * Method used to repaint graph
	 */
	private void repaintGraph()
	{
		remove(graphScrollPane);
		
		createGraph();
		
		graphScrollPane = new JScrollPane(graph);
		
		add(graphScrollPane);
		
		validate();
	}
	
	
	/**
	 * changes zoom of graph
	 * 
	 * @param zoomIn	true to zoom in, false to zoom out
	 */
	private void zoomGraph(boolean zoomIn)
	{
		if (zoomIn)
		{
			graphScale += 0.1;
		}
		else
		{
			graphScale -= 0.1;
		}
		
		if (graphScale > 1.0)
		{
			graphScale = 1.0;
		}
		if (graphScale < 0.1)
		{
			graphScale = 0.1;
		}
		
		graph.setScale(graphScale);
		
		validate();
	}
	
	
	/**
	 * changes distance between nodes
	 * 
	 * @param increaseLevelDistance		true to increase distance, false to decrease instance
	 * 
	 * @param horizontal				true to change horizontal distance, false to change vertical distance
	 */
	public void changeNodesDistance(boolean increaseLevelDistance, boolean horizontal)
	{
		if (horizontal)
		{
			if (increaseLevelDistance)
			{
				horizontalNodesDistance += 20;
			}
			else
			{
				horizontalNodesDistance -= 20;
			}
			
			if (horizontalNodesDistance < 40)
			{
				horizontalNodesDistance = 40;
			}
		}
		else
		{
			if (increaseLevelDistance)
			{
				verticalNodesDistance += 20;
			}
			else
			{
				verticalNodesDistance -= 20;
			}
			
			if (verticalNodesDistance < 40)
			{
				verticalNodesDistance = 40;
			}
		}
		
		
		
		JGraphFacade facade = new JGraphFacade(graph);
		
		JGraphLayout layout = null;
		
		if (treeLayout)
		{
		
			JGraphHierarchicalLayout graphHierarchicalLayout = new JGraphHierarchicalLayout();
			
			graphHierarchicalLayout.setOrientation(graphOrientation);
			graphHierarchicalLayout.setFineTuning(true);
			graphHierarchicalLayout.setFixRoots(true);
			graphHierarchicalLayout.setInterRankCellSpacing(horizontalNodesDistance);
			graphHierarchicalLayout.setIntraCellSpacing(verticalNodesDistance);
			
			layout = graphHierarchicalLayout;
		}
		else
		{
			JGraphFastOrganicLayout graphOrganicLayout = new JGraphFastOrganicLayout();
			
			graphOrganicLayout.setForceConstant(verticalNodesDistance);
			
			layout = graphOrganicLayout;
		}
		
		layout.run(facade);
	
		Map nested = facade.createNestedMap(true, true);
	
		graph.getGraphLayoutCache().edit(nested);
	
		cds.jlow.client.graph.GraphJ.move(graph, cells, 50, 50);
		
		validate();
	}
	
	
	/**
	 * Optimizes graph's size
	 */
	public void optimizeGraphSize()
	{
		double panelHeight = graphScrollPane.getHeight();
		double panelWidth = graphScrollPane.getWidth();
		
		double graphHeight = graph.getPreferredSize().getHeight() / graphScale;
		double graphWidth = graph.getPreferredSize().getWidth() / graphScale;
		
		double verticalScale = panelHeight / (graphHeight+ 50);
		double horizontalScale = panelWidth / (graphWidth + 50);
		
		double scale = verticalScale;
		if (horizontalScale < verticalScale) scale = horizontalScale;
		
		graph.setScale(scale);
		
		graphScale = scale;
		
		graph.repaint();
		
		remove(graphScrollPane);
		
		graphScrollPane = new JScrollPane(graph);
		
		add(graphScrollPane, BorderLayout.CENTER);
		
		validate();
		
		graph.setMinimumSize(new Dimension(graph.getWidth() + arranger.getARDHorizontalNodesDistance(), graph.getHeight()));
	}
	
	
	/**
	 * Rotates graph
	 * 
	 * @param rotateRight		true to rotate right, false to rotate left
	 */
	public void rotateGraph(boolean rotateRight)
	{
		if (rotateRight)
		{
			if (graphOrientation == SwingConstants.WEST) graphOrientation = SwingConstants.NORTH;
			else if (graphOrientation == SwingConstants.NORTH) graphOrientation = SwingConstants.EAST;
			else if (graphOrientation == SwingConstants.EAST) graphOrientation = SwingConstants.SOUTH;
			else if (graphOrientation == SwingConstants.SOUTH) graphOrientation = SwingConstants.WEST;
		}
		else
		{
			if (graphOrientation == SwingConstants.NORTH) graphOrientation = SwingConstants.WEST;
			else if (graphOrientation == SwingConstants.EAST) graphOrientation = SwingConstants.NORTH;
			else if (graphOrientation == SwingConstants.SOUTH) graphOrientation = SwingConstants.EAST;
			else if (graphOrientation == SwingConstants.WEST) graphOrientation = SwingConstants.SOUTH;
		}
		
		repaintGraph();
	}
	
	
	/**
	 * Used to drag graph
	 * 
	 * @param x		x coordinate
	 * @param y		y coordinate
	 */
	private void dragGraph(int x, int y)
	{
		graphScrollPane.getHorizontalScrollBar().setValue(graphScrollPane.getHorizontalScrollBar().getValue() - x);
		graphScrollPane.getVerticalScrollBar().setValue(graphScrollPane.getVerticalScrollBar().getValue() - y);
	}
	
	
	/**
	 * Used to show dialog with given title
	 * 
	 * @param title	title of dialog
	 */
	public void showDialog(String title)
	{	
		dialog = new JFrame(title);
		
		JPanel toolBarsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
	
		
		JToolBar layoutTypeToolBar = new JToolBar();
		
		Action treeLayoutAction = new TreeLayoutAction("Tree layout", arranger.getIcon("tree_layout"));
		Action organicLayoutAction = new OrganicLayoutAction("Organic layout", arranger.getIcon("organic_layout"));
		
		layoutTypeToolBar.add(treeLayoutAction);
		layoutTypeToolBar.add(organicLayoutAction);
		
		layoutTypeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		layoutTypeToolBar.setName("Graph layout");
		
		toolBarsPanel.add(layoutTypeToolBar);
		
		
		JToolBar rotateGraphToolBar = new JToolBar();
		
		Action rotateRightGraphAction = new RotateGraphAction("Rotate right", arranger.getIcon("rotate_right"), true);
		Action rotateLeftGraphAction = new RotateGraphAction("Rotate left", arranger.getIcon("rotate_left"), false);
		
		rotateGraphToolBar.add(rotateRightGraphAction);
		rotateGraphToolBar.add(rotateLeftGraphAction);
		
		rotateGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		rotateGraphToolBar.setName("Rotate graph");
		
		toolBarsPanel.add(rotateGraphToolBar);
		
		
		JToolBar zoomGraphToolBar = new JToolBar();
		
		Action zoomInGraphAction = new ZoomGraphAction("Zoom in", arranger.getIcon("zoom_in"), true);
		Action zoomOutGraphAction = new ZoomGraphAction("Zoom out", arranger.getIcon("zoom_out"), false);
		
		zoomGraphToolBar.add(zoomInGraphAction);
		zoomGraphToolBar.add(zoomOutGraphAction);
		
		zoomGraphToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		zoomGraphToolBar.setName("Zoom graph");
		
		toolBarsPanel.add(zoomGraphToolBar);
		
		
		JToolBar changeDistanceToolBar = new JToolBar();
		
		Action increaseHorizontalDistanceAction = new ChangeNodesDistanceAction("Increase level distance", arranger.getIcon("increase_horizontal_distance"), "horizontal", true);
		Action decreaseHorizontalDistanceAction = new ChangeNodesDistanceAction("Decrease level distance", arranger.getIcon("decrease_horizontal_distance"), "horizontal", false);
		Action increaseVerticalDistanceAction = new ChangeNodesDistanceAction("Increase nodes distance", arranger.getIcon("increase_vertical_distance"), "vertical", true);
		Action decreaseVerticalDistanceAction = new ChangeNodesDistanceAction("Decrease nodes distance", arranger.getIcon("decrease_vertical_distance"), "vertical", false);
		
		changeDistanceToolBar.add(increaseHorizontalDistanceAction);
		changeDistanceToolBar.add(decreaseHorizontalDistanceAction);
		changeDistanceToolBar.add(increaseVerticalDistanceAction);
		changeDistanceToolBar.add(decreaseVerticalDistanceAction);
		
		changeDistanceToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		changeDistanceToolBar.setName("Change nodes distance");
		
		toolBarsPanel.add(changeDistanceToolBar);
		
	
		JToolBar optimizeDiagramSizeToolBar = new JToolBar();
		
		Action optimizeSizeAction = new OptimizeDiagramSizeAction("Optimize diagram size", arranger.getIcon("optimize_graph_size"));
		
		optimizeDiagramSizeToolBar.add(optimizeSizeAction);
		
		optimizeDiagramSizeToolBar.setOrientation(SwingConstants.HORIZONTAL);
		
		optimizeDiagramSizeToolBar.setName("Optimize diagram size");
		
		toolBarsPanel.add(optimizeDiagramSizeToolBar);
		
		
		dialog.add(toolBarsPanel, BorderLayout.NORTH);
		
		dialog.getContentPane().add(this);
		
		dialog.pack();
		
		dialog.setBounds(mainFrame.getX() + 50, mainFrame.getY() + 50, mainFrame.getWidth() - 100, mainFrame.getHeight() - 100);
		dialog.setMinimumSize(new Dimension(300, 300));
		
		dialog.setVisible(true);
	}
	
	
	/**
	 * used to set tree layout
	 * 
	 * @author Szymon Książek
	 */
	private class TreeLayoutAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public TreeLayoutAction(String name, Icon icon)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
		}
		
		public void actionPerformed(ActionEvent event)
		{
			treeLayout = true;
			
			repaintGraph();
		}
	}
	
	
	/**
	 * Used to set organic layout
	 * 
	 * @author Szymon Książek
	 */
	private class OrganicLayoutAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public OrganicLayoutAction(String name, Icon icon)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
		}
		
		public void actionPerformed(ActionEvent event)
		{
			treeLayout = false;
			
			repaintGraph();
		}
	}
	
	
	/**
	 * Used to rotate graph
	 * 
	 * @author Szymon Książek
	 */
	private class RotateGraphAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;
		
		private boolean rotateRight = false;

		public RotateGraphAction(String name, Icon icon, boolean rotateRight)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
			
			this.rotateRight = rotateRight;
		}
		
		public void actionPerformed(ActionEvent event)
		{
			rotateGraph(rotateRight);
		}
	}
	
	
	/**
	 * Used to zoom graph
	 * 
	 * @author Szymon Książek
	 */
	private class ZoomGraphAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;
		
		private boolean zoomIn = false;

		public ZoomGraphAction(String name, Icon icon, boolean zoomIn)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
			
			this.zoomIn = zoomIn;
		}
		
		public void actionPerformed(ActionEvent event)
		{
			zoomGraph(zoomIn);
		}
	}
	
	
	/**
	 * Used to change distance between nodes
	 * 
	 * @author Szymon Książek
	 */
	private class ChangeNodesDistanceAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;
		
		private String distanceType = "";
		private boolean increaseDistance = false;

		public ChangeNodesDistanceAction(String name, Icon icon, String distanceType, boolean increaseDistance)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
			
			this.distanceType = distanceType;
			this.increaseDistance = increaseDistance;
		}
		
		public void actionPerformed(ActionEvent event)
		{
			boolean horizontal = true;
			
			if (distanceType.equals("vertical")) horizontal = false;
			
			changeNodesDistance(increaseDistance, horizontal);
		}
	}
	
	
	/**
	 * Used to optimize graph's size
	 * 
	 * @author Szymon Książek
	 */
	private class OptimizeDiagramSizeAction extends AbstractAction
	{
		private static final long serialVersionUID = 1L;

		public OptimizeDiagramSizeAction(String name, Icon icon)
		{
			putValue(Action.NAME, name);
			putValue(Action.SMALL_ICON, icon);
			putValue(Action.SHORT_DESCRIPTION, name);
		}
		
		public void actionPerformed(ActionEvent event)
		{
			optimizeGraphSize();
		}
	}
	
	
	/**
	 * Router that sets spline style of edges
	 * 
	 * @author Szymon Książek
	 */
	private class EdgeRouter extends DefaultEdge.LoopRouting 
	{
		private static final long serialVersionUID = 1L;

		public int getEdgeStyle() 
		{
			return GraphConstants.STYLE_SPLINE;
		}
	}
	
	
	/**
	 * Mouse listener
	 * 
	 * @author Szymon Książek
	 */
	public class MouseListener extends MouseAdapter 
	{
		/**
		 * Specifies last coordinates of dragged point
		 */
		private Point dragStartPoint = null;
		
		
		/**
		 * Invoked when a mouse button has been double pressed on graph 
		 */
		public void mousePressed(MouseEvent e)
		{
			if (e.isAltDown())
			{	
				dragStartPoint = e.getLocationOnScreen();
			}
		}
		
		/**
		 * invoked when graph is dragged
		 */
		public void mouseDragged(MouseEvent e)
		{
			if (e.isAltDown())
			{
				Point actualPoint = e.getLocationOnScreen();
				
				int xMove = (int)(actualPoint.getX() - dragStartPoint.getX());
				int yMove = (int)(actualPoint.getY() - dragStartPoint.getY());
				
				if (xMove != 0 || yMove != 0)
				{
					dragGraph(xMove, yMove);
				}
				
				dragStartPoint = actualPoint;
			}
		}
		
		
		public void mouseWheelMoved(MouseWheelEvent e)
		{
			// number of 'clicks' - how fast wheel is rotated
			int clicks = e.getWheelRotation();
			
			// zoom out graph
			if (clicks > 0)
			{
				for (int zoomCount = 0 ; zoomCount < clicks ; zoomCount++)
				{
					zoomGraph(false);
				}
			}
			// zoom in graph
			else if (clicks < 0)
			{
				clicks = -clicks;
				
				for (int zoomCount = 0 ; zoomCount < clicks ; zoomCount++)
				{
					zoomGraph(true);
				}
			}
		}
	}
}